"""Simulation of a block connected to a spring in a viscous medium"""
from vedo import *

L = 0.1    # spring x position at rest
x0 = 0.85  # initial x-coordinate of the block
k = 25     # spring constant
m = 20     # block mass
b = 0.5    # viscosity friction (proportional to velocity)
dt = 0.15  # time step

v = vector(0, 0, 0.2)  # initial conditions
x = vector(x0, 0, 0)
x0 = vector(-0.8, 0, 0)
xr = vector(L, 0, 0)

def loop_func(event):
    global v, x
    F = -k * (x - xr) - b * v       # force and friction
    a = F / m                       # acceleration
    v = v + a*dt                    # velocity
    x = x + v*dt + 1/2 * a * dt**2  # position

    block.pos(x)  # update block position and trail
    spring.stretch(x0, x)  # stretch helix accordingly
    plt.render()

block = Cube(pos=x, side=0.2, c="tomato")
spring = Spring(x0, x, r1=0.06, thickness=0.01)

plt = Plotter(size=(1050, 600))
plt += Box(pos=(0, -0.1, 0), size=(2.0, 0.02, 0.5))    # floor
plt += Box(pos=(-0.82, 0.15, 0), size=(0.04,0.50,0.3)) # wall
plt += [block, spring, __doc__]
plt.add_callback("timer", loop_func)
plt.timer_callback("start")
plt.show().close()



from vedo import *

# Enable depth peeling for rendering transparency
settings.use_depth_peeling = True

# Declare an instance of the Plotter class with 2 rows and 2 columns of renderers,
# and disable interactive mode, so that the program can continue running
plt = Plotter(shape=(2, 2), interactive=False, axes=3)

# Create two sphere meshes
s1 = Sphere(pos=[-0.7, 0, 0], c="red5", alpha=0.5)
s2 = Sphere(pos=[0.7, 0, 0], c="green5", alpha=0.5)

# Show the spheres on the first renderer, and display the docstring as the title
plt.at(0).show(s1, s2, __doc__)

# Perform a boolean intersection operation between the two spheres,
# set the color to magenta, and show the result on the second renderer
b1 = s1.boolean("intersect", s2).c('magenta')
plt.at(1).show(b1, "intersect", resetcam=False)

# Perform a boolean union operation between the two spheres,
# set the color to blue, add a wireframe, and show the result on the third renderer
b2 = s1.boolean("plus", s2).c("blue").wireframe(True)
plt.at(2).show(b2, "plus", resetcam=False)

# Perform a boolean difference operation between the two spheres,
# compute the normals, add a scalarbar, and show the result on the fourth renderer
b3 = s1.boolean("minus", s2).compute_normals().add_scalarbar(c='white')
plt.at(3).show(b3, "minus", resetcam=False)

# Enable interactive mode, and close the plot
plt.interactive().close()



import numpy as np

try:
    import vedo.vtkclasses as vtk
except ImportError:
    import vtkmodules.all as vtk

import vedo
from vedo import settings
from vedo import colors
from vedo import utils
from vedo.base import BaseGrid
from vedo.file_io import download, loadUnStructuredGrid


__docformat__ = "google"

__doc__ = """
Work with unstructured grid datasets
"""

__all__ = ["UGrid"]

#########################################################################
class UGrid(BaseGrid, vtk.vtkActor):
    """Support for UnstructuredGrid objects."""

    def __init__(self, inputobj=None):
        """
        Support for UnstructuredGrid objects.

        Arguments:
            inputobj : (list, vtkUnstructuredGrid, str)
                A list in the form `[points, cells, celltypes]`,
                or a vtkUnstructuredGrid object, or a filename

        Celltypes are identified by the following convention:
            - VTK_TETRA = 10
            - VTK_VOXEL = 11
            - VTK_HEXAHEDRON = 12
            - VTK_WEDGE = 13
            - VTK_PYRAMID = 14
            - VTK_HEXAGONAL_PRISM = 15
            - VTK_PENTAGONAL_PRISM = 16
        """

        vtk.vtkActor.__init__(self)
        BaseGrid.__init__(self)

        inputtype = str(type(inputobj))
        self._data = None
        self._polydata = None
        self._bfprop = None
        self.name = "UGrid"

        ###################
        if inputobj is None:
            self._data = vtk.vtkUnstructuredGrid()

        elif utils.is_sequence(inputobj):

            pts, cells, celltypes = inputobj

            self._data = vtk.vtkUnstructuredGrid()

            if not utils.is_sequence(cells[0]):
                tets = []
                nf = cells[0] + 1
                for i, cl in enumerate(cells):
                    if i in (nf, 0):
                        k = i + 1
                        nf = cl + k
                        cell = [cells[j + k] for j in range(cl)]
                        tets.append(cell)
                cells = tets

            # This would fill the points and use those to define orientation
            vpts = utils.numpy2vtk(pts, dtype=np.float32)
            points = vtk.vtkPoints()
            points.SetData(vpts)
            self._data.SetPoints(points)

            # This fill the points and use cells to define orientation
            # points = vtk.vtkPoints()
            # for c in cells:
            #       for pid in c:
            #           points.InsertNextPoint(pts[pid])
            # self._data.SetPoints(points)

            # Fill cells
            # https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html
            for i, ct in enumerate(celltypes):
                cell_conn = cells[i]
                if ct == vtk.VTK_HEXAHEDRON:
                    cell = vtk.vtkHexahedron()
                elif ct == vtk.VTK_TETRA:
                    cell = vtk.vtkTetra()
                elif ct == vtk.VTK_VOXEL:
                    cell = vtk.vtkVoxel()
                elif ct == vtk.VTK_WEDGE:
                    cell = vtk.vtkWedge()
                elif ct == vtk.VTK_PYRAMID:
                    cell = vtk.vtkPyramid()
                elif ct == vtk.VTK_HEXAGONAL_PRISM:
                    cell = vtk.vtkHexagonalPrism()
                elif ct == vtk.VTK_PENTAGONAL_PRISM:
                    cell = vtk.vtkPentagonalPrism()
                else:
                    print("UGrid: cell type", ct, "not implemented. Skip.")
                    continue
                cpids = cell.GetPointIds()
                for j, pid in enumerate(cell_conn):
                    cpids.SetId(j, pid)
                self._data.InsertNextCell(ct, cpids)

        elif "UnstructuredGrid" in inputtype:
            self._data = inputobj

        elif isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = download(inputobj, verbose=False)
            self._data = loadUnStructuredGrid(inputobj)
            self.filename = inputobj

        else:
            vedo.logger.error(f"cannot understand input type {inputtype}")
            return

        # self._mapper = vtk.vtkDataSetMapper()
        self._mapper = vtk.vtkPolyDataMapper()
        self._mapper.SetInterpolateScalarsBeforeMapping(settings.interpolate_scalars_before_mapping)

        if settings.use_polygon_offset:
            self._mapper.SetResolveCoincidentTopologyToPolygonOffset()
            pof, pou = settings.polygon_offset_factor, settings.polygon_offset_units
            self._mapper.SetResolveCoincidentTopologyPolygonOffsetParameters(pof, pou)
        self.GetProperty().SetInterpolationToFlat()

        if not self._data:
            return

        # now fill the representation of the vtk unstr grid
        sf = vtk.vtkShrinkFilter()
        sf.SetInputData(self._data)
        sf.SetShrinkFactor(1.0)
        sf.Update()
        gf = vtk.vtkGeometryFilter()
        gf.SetInputData(sf.GetOutput())
        gf.Update()
        self._polydata = gf.GetOutput()

        self._mapper.SetInputData(self._polydata)
        sc = None
        if self.useCells:
            sc = self._polydata.GetCellData().GetScalars()
        else:
            sc = self._polydata.GetPointData().GetScalars()
        if sc:
            self._mapper.SetScalarRange(sc.GetRange())

        self.SetMapper(self._mapper)
        self.property = self.GetProperty()

        self.pipeline = utils.OperationNode(
            self, comment=f"#cells {self._data.GetNumberOfCells()}",
            c="#4cc9f0",
        )
    # ------------------------------------------------------------------

    def _repr_html_(self):
        """
        HTML representation of the UGrid object for Jupyter Notebooks.

        Returns:
            HTML text with the image and some properties.
        """
        import io
        import base64
        from PIL import Image

        library_name = "vedo.ugrid.UGrid"
        help_url = "https://vedo.embl.es/docs/vedo/ugrid.html"

        arr = self.thumbnail()
        im = Image.fromarray(arr)
        buffered = io.BytesIO()
        im.save(buffered, format="PNG", quality=100)
        encoded = base64.b64encode(buffered.getvalue()).decode("utf-8")
        url = "data:image/png;base64," + encoded
        image = f"<img src='{url}'></img>"

        bounds = "<br/>".join(
            [
                utils.precision(min_x,4) + " ... " + utils.precision(max_x,4)
                for min_x, max_x in zip(self.bounds()[::2], self.bounds()[1::2])
            ]
        )

        help_text = ""
        if self.name:
            help_text += f"<b> {self.name}: &nbsp&nbsp</b>"
        help_text += '<b><a href="' + help_url + '" target="_blank">' + library_name + "</a></b>"
        if self.filename:
            dots = ""
            if len(self.filename) > 30:
                dots = "..."
            help_text += f"<br/><code><i>({dots}{self.filename[-30:]})</i></code>"

        pdata = ""
        if self._data.GetPointData().GetScalars():
            if self._data.GetPointData().GetScalars().GetName():
                name = self._data.GetPointData().GetScalars().GetName()
                pdata = "<tr><td><b> point data array </b></td><td>" + name + "</td></tr>"

        cdata = ""
        if self._data.GetCellData().GetScalars():
            if self._data.GetCellData().GetScalars().GetName():
                name = self._data.GetCellData().GetScalars().GetName()
                cdata = "<tr><td><b> cell data array </b></td><td>" + name + "</td></tr>"

        pts = self.points()
        cm = np.mean(pts, axis=0)

        all = [
            "<table>",
            "<tr>",
            "<td>", image, "</td>",
            "<td style='text-align: center; vertical-align: center;'><br/>", help_text,
            "<table>",
            "<tr><td><b> bounds </b> <br/> (x/y/z) </td><td>" + str(bounds) + "</td></tr>",
            "<tr><td><b> center of mass </b></td><td>" + utils.precision(cm,3) + "</td></tr>",
            # "<tr><td><b> average size </b></td><td>" + str(average_size) + "</td></tr>",
            "<tr><td><b> nr. points&nbsp/&nbspcells </b></td><td>"
            + str(self.npoints) + "&nbsp/&nbsp" + str(self.ncells) + "</td></tr>",
            pdata,
            cdata,
            "</table>",
            "</table>",
        ]
        return "\n".join(all)

    def clone(self):
        """Clone the UGrid object to yield an exact copy."""
        ugCopy = vtk.vtkUnstructuredGrid()
        ugCopy.DeepCopy(self._data)

        cloned = UGrid(ugCopy)
        pr = self.GetProperty()
        if isinstance(pr, vtk.vtkVolumeProperty):
            prv = vtk.vtkVolumeProperty()
        else:
            prv = vtk.vtkProperty()
        prv.DeepCopy(pr)
        cloned.SetProperty(prv)
        cloned.property = prv

        # assign the same transformation to the copy
        cloned.SetOrigin(self.GetOrigin())
        cloned.SetScale(self.GetScale())
        cloned.SetOrientation(self.GetOrientation())
        cloned.SetPosition(self.GetPosition())
        cloned.name = self.name

        cloned.pipeline = utils.OperationNode(
            "clone", parents=[self], shape='diamond', c='#bbe1ed',
        )
        return cloned

    def color(self, c=False, alpha=None):
        """
        Set/get UGrid color.
        If None is passed as input, will use colors from active scalars.
        Same as `ugrid.c()`.
        """
        if c is False:
            return np.array(self.GetProperty().GetColor())
        if c is None:
            self._mapper.ScalarVisibilityOn()
            return self
        self._mapper.ScalarVisibilityOff()
        cc = colors.get_color(c)
        self.property.SetColor(cc)
        if self.trail:
            self.trail.GetProperty().SetColor(cc)
        if alpha is not None:
            self.alpha(alpha)
        return self

    def alpha(self, opacity=None):
        """Set/get mesh's transparency. Same as `mesh.opacity()`."""
        if opacity is None:
            return self.property.GetOpacity()

        self.property.SetOpacity(opacity)
        bfp = self.GetBackfaceProperty()
        if bfp:
            if opacity < 1:
                self._bfprop = bfp
                self.SetBackfaceProperty(None)
            else:
                self.SetBackfaceProperty(self._bfprop)
        return self

    def opacity(self, alpha=None):
        """Set/get mesh's transparency. Same as `mesh.alpha()`."""
        return self.alpha(alpha)

    def wireframe(self, value=True):
        """Set mesh's representation as wireframe or solid surface.
        Same as `mesh.wireframe()`."""
        if value:
            self.property.SetRepresentationToWireframe()
        else:
            self.property.SetRepresentationToSurface()
        return self

    def linewidth(self, lw=None):
        """Set/get width of mesh edges. Same as `lw()`."""
        if lw is not None:
            if lw == 0:
                self.property.EdgeVisibilityOff()
                self.property.SetRepresentationToSurface()
                return self
            self.property.EdgeVisibilityOn()
            self.property.SetLineWidth(lw)
        else:
            return self.property.GetLineWidth()
        return self

    def lw(self, linewidth=None):
        """Set/get width of mesh edges. Same as `linewidth()`."""
        return self.linewidth(linewidth)

    def linecolor(self, lc=None):
        """Set/get color of mesh edges. Same as `lc()`."""
        if lc is not None:
            if "ireframe" in self.property.GetRepresentationAsString():
                self.property.EdgeVisibilityOff()
                self.color(lc)
                return self
            self.property.EdgeVisibilityOn()
            self.property.SetEdgeColor(colors.get_color(lc))
        else:
            return self.property.GetEdgeColor()
        return self

    def lc(self, linecolor=None):
        """Set/get color of mesh edges. Same as `linecolor()`."""
        return self.linecolor(linecolor)

    def extract_cell_type(self, ctype):
        """Extract a specific cell type and return a new `UGrid`."""
        uarr = self._data.GetCellTypesArray()
        ctarrtyp = np.where(utils.vtk2numpy(uarr) == ctype)[0]
        uarrtyp = utils.numpy2vtk(ctarrtyp, deep=False, dtype="id")
        selection_node = vtk.vtkSelectionNode()
        selection_node.SetFieldType(vtk.vtkSelectionNode.CELL)
        selection_node.SetContentType(vtk.vtkSelectionNode.INDICES)
        selection_node.SetSelectionList(uarrtyp)
        selection = vtk.vtkSelection()
        selection.AddNode(selection_node)
        es = vtk.vtkExtractSelection()
        es.SetInputData(0, self._data)
        es.SetInputData(1, selection)
        es.Update()
        ug = UGrid(es.GetOutput())

        ug.pipeline = utils.OperationNode(
            "extract_cell_type", comment=f"type {ctype}",
            c="#edabab", parents=[self],
        )
        return ug


"""Build a custom colormap, including
out-of-range and NaN colors and labels"""
from vedo import build_lut, Sphere, show, settings

settings.use_depth_peeling = True # might help with transparencies

# generate a sphere and stretch it, so it sits between z=-2 and z=+2
mesh = Sphere(quads=True).scale([1,1,2]).linewidth(0.1)

# create some dummy data array to be associated to points
data = mesh.points()[:,2]  # pick z-coords, use them as scalar data
data[10:70] = float('nan') # make some values invalid by setting to NaN
data[300:600] = 100        # send some values very far above-scale

# build a custom Look-Up-Table of colors:
#     value, color, alpha
lut = build_lut(
    [
      #(-2, 'pink'      ),  # up to -2 is pink
      (0.0, 'pink'      ),  # up to 0 is pink
      (0.4, 'green', 0.5),  # up to 0.4 is green with alpha=0.5
      (0.7, 'darkblue'  ),
      #( 2, 'darkblue'  ),
    ],
    vmin=-1.2,
    vmax= 0.7,
    below_color='lightblue',
    above_color='grey',
    nan_color='red',
    interpolate=False,
)
# 3D scalarbar:
mesh.cmap(lut, data).add_scalarbar3d(title='My 3D scalarbar', c='white')
mesh.scalarbar.scale(1.5).rotate_x(90).y(1) # make it bigger and place it

# 2D scalarbar:
# mesh.cmap(lut, data).add_scalarbar()

show(mesh, __doc__,
     axes=dict(zlabel_size=.04, number_of_divisions=10),
     elevation=-80, bg='blackboard',
).close()



"""Superimpose and compare histograms"""
import numpy as np
from vedo.pyplot import histogram
from vedo import settings

settings.remember_last_figure_format = True

np.random.seed(0)
theory = np.random.randn(500).tolist()
backgr = ((np.random.rand(100)-0.5)*6).tolist()
data = np.random.randn(500).tolist() + backgr

# A first histogram:
fig = histogram(
    theory + backgr,
    ylim=(0,90),
    title=__doc__,
    xtitle='measured variable',
    c='red4',
    gap=0,      # no gap between bins
    padding=0,  # no extra spaces
    label="theory",
)

# Extract the 11th bin and color it purple
fig.unpack(10).c('purple4')
fig.add_label("special bin", marker='s', mc='purple4')

# Add a second histogram to be superimposed
fig += histogram(backgr, label='background')

# Add the data histogram with poissonian errors
fig += histogram(data, marker='o', errors=True, fill=False, label='data')

fig.add_legend(s=0.8)
fig.show(zoom='tight').close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Subset of vtk classes to be imported directly
"""

import vtkmodules.vtkCommonComputationalGeometry

from vtkmodules.vtkCommonColor import vtkNamedColors

from vtkmodules.vtkCommonCore import (
    mutable,
    VTK_UNSIGNED_CHAR,
    VTK_UNSIGNED_SHORT,
    VTK_UNSIGNED_INT,
    VTK_UNSIGNED_LONG,
    VTK_UNSIGNED_LONG_LONG,
    VTK_UNSIGNED_CHAR,
    VTK_CHAR,
    VTK_SHORT,
    VTK_INT,
    VTK_LONG,
    VTK_LONG_LONG,
    VTK_FLOAT,
    VTK_DOUBLE,
    VTK_SIGNED_CHAR,
    VTK_ID_TYPE,
    VTK_VERSION_NUMBER,
    VTK_FONT_FILE,
    vtkArray,
    vtkIdTypeArray,
    vtkBitArray,
    vtkCharArray,
    vtkDoubleArray,
    vtkFloatArray,
    vtkIdList,
    vtkIntArray,
    vtkLookupTable,
    vtkMath,
    vtkPoints,
    vtkStringArray,
    vtkUnsignedCharArray,
    vtkVariant,
    vtkVariantArray,
    vtkVersion,
)

from vtkmodules.vtkCommonDataModel import (
    VTK_HEXAHEDRON,
    VTK_TETRA,
    VTK_VOXEL,
    VTK_WEDGE,
    VTK_PYRAMID,
    VTK_HEXAGONAL_PRISM,
    VTK_PENTAGONAL_PRISM,
    vtkCellArray,
    vtkBox,
    vtkCellLocator,
    vtkCylinder,
    vtkDataSetAttributes,
    vtkDataObject,
    vtkDataSet,
    vtkFieldData,
    vtkHexagonalPrism,
    vtkHexahedron,
    vtkImageData,
    vtkImplicitDataSet,
    vtkImplicitSelectionLoop,
    vtkImplicitWindowFunction,
    vtkIterativeClosestPointTransform,
    vtkLine,
    vtkMultiBlockDataSet,
    vtkMutableDirectedGraph,
    vtkPentagonalPrism,
    vtkPlane,
    vtkPlanes,
    vtkPointLocator,
    vtkPolyData,
    vtkPolyLine,
    vtkPolyPlane,
    vtkPolygon,
    vtkPyramid,
    vtkQuadric,
    vtkRectilinearGrid,
    vtkSelection,
    vtkSelectionNode,
    vtkSphere,
    vtkStaticCellLocator,
    vtkStaticPointLocator,
    vtkStructuredGrid,
    vtkTetra,
    vtkTriangle,
    vtkUnstructuredGrid,
    vtkVoxel,
    vtkWedge,
)

from vtkmodules.vtkCommonExecutionModel import vtkAlgorithm

from vtkmodules.vtkCommonMath import vtkMatrix4x4, vtkQuaternion

from vtkmodules.vtkCommonTransforms import (
    vtkHomogeneousTransform,
    vtkLandmarkTransform,
    vtkThinPlateSplineTransform,
    vtkTransform
)

from vtkmodules.vtkFiltersCore import (
    VTK_BEST_FITTING_PLANE,
    vtk3DLinearGridCrinkleExtractor,
    vtkAppendPolyData,
    vtkCellCenters,
    vtkCellDataToPointData,
    vtkCenterOfMass,
    vtkCleanPolyData,
    vtkClipPolyData,
    vtkPolyDataConnectivityFilter,
    vtkPolyDataEdgeConnectivityFilter,
    vtkContourFilter,
    vtkContourGrid,
    vtkCutter,
    vtkDecimatePro,
    vtkDelaunay2D,
    vtkDelaunay3D,
    vtkElevationFilter,
    vtkFeatureEdges,
    vtkFlyingEdges3D,
    vtkGlyph3D,
    vtkIdFilter,
    vtkImageAppend,
    vtkImplicitPolyDataDistance,
    vtkMarchingSquares,
    vtkMaskPoints,
    vtkMassProperties,
    vtkPointDataToCellData,
    vtkPolyDataNormals,
    vtkProbeFilter,
    vtkQuadricDecimation,
    vtkResampleWithDataSet,
    vtkReverseSense,
    vtkStripper,
    vtkTensorGlyph,
    vtkThreshold,
    vtkTriangleFilter,
    vtkTubeFilter,
    vtkUnstructuredGridQuadricDecimation,
    vtkVoronoi2D,
    vtkWindowedSincPolyDataFilter,
)

try:
    from vtkmodules.vtkFiltersCore import vtkStaticCleanUnstructuredGrid, vtkPolyDataPlaneCutter
except ImportError:
    pass


from vtkmodules.vtkFiltersExtraction import (
    vtkExtractCellsByType,
    vtkExtractGeometry,
    vtkExtractPolyDataGeometry,
    vtkExtractSelection,
)
try:
    from vtkmodules.vtkFiltersExtraction import vtkExtractEdges # vtk9.0
except ImportError:
    from vtkmodules.vtkFiltersCore import vtkExtractEdges # vtk9.2

from vtkmodules.vtkFiltersFlowPaths import vtkStreamTracer


from vtkmodules.vtkFiltersGeneral import (
    vtkBooleanOperationPolyDataFilter,
    vtkBoxClipDataSet,
    vtkCellValidator,
    vtkClipDataSet,
    vtkCountVertices,
    vtkContourTriangulator,
    vtkCurvatures,
    vtkDataSetTriangleFilter,
    vtkDensifyPolyData,
    vtkDistancePolyDataFilter,
    vtkGradientFilter,
    vtkIntersectionPolyDataFilter,
    vtkLoopBooleanPolyDataFilter,
    vtkMultiBlockDataGroupFilter,
    vtkTransformPolyDataFilter,
    vtkOBBTree,
    vtkQuantizePolyDataPoints,
    vtkRandomAttributeGenerator,
    vtkShrinkFilter,
    vtkShrinkPolyData,
    vtkRectilinearGridToTetrahedra,
    vtkVertexGlyphFilter,
)

try:
    from vtkmodules.vtkCommonDataModel import vtkCellTreeLocator
except ImportError:
    from vtkmodules.vtkFiltersGeneral import vtkCellTreeLocator

from vtkmodules.vtkFiltersGeometry import (
    vtkGeometryFilter,
    vtkDataSetSurfaceFilter,
    vtkImageDataGeometryFilter,
)
try:
    from vtkmodules.vtkFiltersGeometry import vtkMarkBoundaryFilter
except ImportError:
    pass


from vtkmodules.vtkFiltersHybrid import (
    vtkFacetReader,
    vtkImplicitModeller,
    vtkPolyDataSilhouette,
    vtkProcrustesAlignmentFilter,
    vtkRenderLargeImage,
)

from vtkmodules.vtkFiltersModeling import (
    vtkAdaptiveSubdivisionFilter,
    vtkBandedPolyDataContourFilter,
    vtkButterflySubdivisionFilter,
    vtkContourLoopExtraction,
    vtkCookieCutter,
    vtkDijkstraGraphGeodesicPath,
    vtkFillHolesFilter,
    vtkHausdorffDistancePointSetFilter,
    vtkLinearExtrusionFilter,
    vtkLinearSubdivisionFilter,
    vtkLoopSubdivisionFilter,
    vtkRibbonFilter,
    vtkRotationalExtrusionFilter,
    vtkRuledSurfaceFilter,
    vtkSectorSource,
    vtkSelectEnclosedPoints,
    vtkSelectPolyData,
    vtkSubdivideTetra,
)

try:
    from vtkmodules.vtkFiltersModeling import vtkCollisionDetectionFilter
except ImportError:
    pass

try:
    from vtkmodules.vtkFiltersModeling import vtkImprintFilter
except ImportError:
    pass

from vtkmodules.vtkFiltersPoints import (
    vtkConnectedPointsFilter,
    vtkDensifyPointCloudFilter,
    vtkEuclideanClusterExtraction,
    vtkExtractEnclosedPoints,
    vtkExtractSurface,
    vtkGaussianKernel,
    vtkLinearKernel,
    vtkPCANormalEstimation,
    vtkPointDensityFilter,
    vtkPointInterpolator,
    vtkRadiusOutlierRemoval,
    vtkShepardKernel,
    vtkSignedDistance,
    vtkVoronoiKernel,
)

from vtkmodules.vtkFiltersSources import (
    vtkArcSource,
    vtkArrowSource,
    vtkConeSource,
    vtkCubeSource,
    vtkCylinderSource,
    vtkDiskSource,
    vtkGlyphSource2D,
    vtkGraphToPolyData,
    vtkLineSource,
    vtkOutlineCornerFilter,
    vtkParametricFunctionSource,
    vtkPlaneSource,
    vtkPointSource,
    vtkProgrammableSource,
    vtkSphereSource,
    vtkTexturedSphereSource,
    vtkTessellatedBoxSource,
)


from vtkmodules.vtkFiltersTexture import vtkTextureMapToPlane

from vtkmodules.vtkFiltersVerdict import vtkMeshQuality, vtkCellSizeFilter

from vtkmodules.vtkImagingStencil import vtkPolyDataToImageStencil

from vtkmodules.vtkIOExport import vtkX3DExporter

from vtkmodules.vtkIOExportGL2PS import vtkGL2PSExporter

from vtkmodules.vtkIOGeometry import (
    vtkBYUReader,
    vtkFacetWriter,
    vtkOBJReader,
    vtkOpenFOAMReader,
    vtkParticleReader,
    vtkSTLReader,
    vtkSTLWriter,
)

from vtkmodules.vtkIOImage import (
    vtkBMPReader,
    vtkBMPWriter,
    vtkDEMReader,
    vtkDICOMImageReader,
    vtkHDRReader,
    vtkJPEGReader,
    vtkJPEGWriter,
    vtkMetaImageReader,
    vtkMetaImageWriter,
    vtkNIFTIImageReader,
    vtkNIFTIImageWriter,
    vtkNrrdReader,
    vtkPNGReader,
    vtkPNGWriter,
    vtkSLCReader,
    vtkTIFFReader,
    vtkTIFFWriter,
)

from vtkmodules.vtkIOImport import (
    vtk3DSImporter,
    vtkOBJImporter,
    vtkVRMLImporter,
)

from vtkmodules.vtkIOLegacy import (
    vtkSimplePointsWriter,
    vtkStructuredGridReader,
    vtkStructuredPointsReader,
    vtkDataSetReader,
    vtkDataSetWriter,
    vtkPolyDataWriter,
    vtkRectilinearGridReader,
    vtkUnstructuredGridReader,
)

from vtkmodules.vtkIOPLY import vtkPLYReader, vtkPLYWriter

from vtkmodules.vtkIOXML import (
    vtkXMLGenericDataObjectReader,
    vtkXMLImageDataReader,
    vtkXMLImageDataWriter,
    vtkXMLMultiBlockDataReader,
    vtkXMLMultiBlockDataWriter,
    vtkXMLPRectilinearGridReader,
    vtkXMLPUnstructuredGridReader,
    vtkXMLPolyDataReader,
    vtkXMLPolyDataWriter,
    vtkXMLRectilinearGridReader,
    vtkXMLStructuredGridReader,
    vtkXMLUnstructuredGridReader,
    vtkXMLUnstructuredGridWriter,
)

from vtkmodules.vtkImagingColor import (
    vtkImageLuminance,
    vtkImageMapToWindowLevelColors,
)

from vtkmodules.vtkImagingCore import (
    vtkExtractVOI,
    vtkImageAppendComponents,
    vtkImageBlend,
    vtkImageCast,
    vtkImageConstantPad,
    vtkImageExtractComponents,
    vtkImageFlip,
    vtkImageMapToColors,
    vtkImageMirrorPad,
    vtkImagePermute,
    vtkImageResample,
    vtkImageResize,
    vtkImageReslice,
    vtkImageThreshold,
    vtkImageTranslateExtent,
)

from vtkmodules.vtkImagingFourier import (
    vtkImageButterworthHighPass,
    vtkImageButterworthLowPass,
    vtkImageFFT,
    vtkImageFourierCenter,
    vtkImageRFFT,
)

from vtkmodules.vtkImagingGeneral import (
    vtkImageCorrelation,
    vtkImageEuclideanDistance,
    vtkImageGaussianSmooth,
    vtkImageGradient,
    vtkImageHybridMedian2D,
    vtkImageLaplacian,
    vtkImageMedian3D,
    vtkImageNormalize,
)

from vtkmodules.vtkImagingHybrid import vtkImageToPoints, vtkSampleFunction
from vtkmodules.vtkImagingMath import (
    vtkImageDivergence,
    vtkImageDotProduct,
    vtkImageLogarithmicScale,
    vtkImageMagnitude,
    vtkImageMathematics,
)

from vtkmodules.vtkImagingMorphological import (
    vtkImageContinuousDilate3D,
    vtkImageContinuousErode3D,
)

from vtkmodules.vtkImagingSources import vtkImageCanvasSource2D
from vtkmodules.vtkImagingStencil import vtkImageStencil
from vtkmodules.vtkInfovisLayout import (
    vtkCircularLayoutStrategy,
    vtkClustering2DLayoutStrategy,
    vtkConeLayoutStrategy,
    vtkFast2DLayoutStrategy,
    vtkForceDirectedLayoutStrategy,
    vtkGraphLayout,
    vtkSimple2DLayoutStrategy,
    vtkSimple3DCirclesStrategy,
    vtkSpanTreeLayoutStrategy,
)

from vtkmodules.vtkInteractionStyle import (
    vtkInteractorStyleFlight,
    vtkInteractorStyleImage,
    vtkInteractorStyleJoystickActor,
    vtkInteractorStyleJoystickCamera,
    vtkInteractorStyleRubberBand2D,
    vtkInteractorStyleRubberBand3D,
    vtkInteractorStyleRubberBandZoom,
    vtkInteractorStyleTerrain,
    vtkInteractorStyleTrackballActor,
    vtkInteractorStyleTrackballCamera,
    vtkInteractorStyleUnicam,
    vtkInteractorStyleUser,
)

from vtkmodules.vtkInteractionWidgets import (
    vtkBalloonRepresentation,
    vtkBalloonWidget,
    vtkBoxWidget,
    vtkContourWidget,
    vtkPlaneWidget,
    vtkFocalPlanePointPlacer,
    vtkImplicitPlaneWidget,
    vtkOrientationMarkerWidget,
    vtkOrientedGlyphContourRepresentation,
    vtkPolygonalSurfacePointPlacer,
    vtkSliderRepresentation2D,
    vtkSliderRepresentation3D,
    vtkSliderWidget,
    vtkSphereWidget,
)

try:
    from vtkmodules.vtkInteractionWidgets import vtkCameraOrientationWidget
except ImportError:
    pass

from vtkmodules.vtkRenderingAnnotation import (
    vtkAnnotatedCubeActor,
    vtkAxesActor,
    vtkAxisActor2D,
    vtkCaptionActor2D,
    vtkCornerAnnotation,
    vtkCubeAxesActor,
    vtkLegendBoxActor,
    vtkLegendScaleActor,
    vtkPolarAxesActor,
    vtkScalarBarActor,
    vtkXYPlotActor,
)

from vtkmodules.vtkRenderingCore import (
    vtkActor,
    vtkActor2D,
    vtkAssembly,
    vtkBillboardTextActor3D,
    vtkCamera,
    vtkCameraInterpolator,
    vtkColorTransferFunction,
    vtkCoordinate,
    vtkDataSetMapper,
    vtkDistanceToCamera,
    vtkFlagpoleLabel,
    vtkFollower,
    vtkHierarchicalPolyDataMapper,
    vtkImageActor,
    vtkImageMapper,
    vtkImageProperty,
    vtkImageSlice,
    vtkInteractorEventRecorder,
    vtkInteractorObserver,
    vtkLight,
    vtkLogLookupTable,
    vtkMapper,
    vtkPointGaussianMapper,
    vtkPolyDataMapper,
    vtkPolyDataMapper2D,
    vtkProp,
    vtkProp3D,
    vtkPropAssembly,
    vtkPropCollection,
    vtkPropPicker,
    vtkProperty,
    vtkRenderWindow,
    vtkRenderer,
    vtkRenderWindowInteractor,
    vtkSelectVisiblePoints,
    vtkSkybox,
    vtkTextActor,
    vtkTextMapper,
    vtkTextProperty,
    vtkTextRenderer,
    vtkTexture,
    vtkVolume,
    vtkVolumeProperty,
    vtkWindowToImageFilter,
)

from vtkmodules.vtkRenderingFreeType import vtkVectorText

from vtkmodules.vtkRenderingImage import vtkImageResliceMapper

from vtkmodules.vtkRenderingLabel import vtkLabeledDataMapper

from vtkmodules.vtkRenderingOpenGL2 import (
    vtkDepthOfFieldPass,
    vtkCameraPass,
    vtkDualDepthPeelingPass,
    vtkEquirectangularToCubeMapTexture,
    vtkLightsPass,
    vtkOpaquePass,
    vtkOverlayPass,
    vtkRenderPassCollection,
    vtkSSAOPass,
    vtkSequencePass,
    vtkShader,
    vtkShadowMapPass,
    vtkTranslucentPass,
    vtkVolumetricPass,
)

from vtkmodules.vtkRenderingVolume import (
    vtkFixedPointVolumeRayCastMapper,
    vtkGPUVolumeRayCastMapper,
    vtkProjectedTetrahedraMapper,
    vtkUnstructuredGridVolumeRayCastMapper,
    vtkUnstructuredGridVolumeZSweepMapper,
)

from vtkmodules.vtkRenderingVolumeOpenGL2 import (
    vtkOpenGLGPUVolumeRayCastMapper,
    vtkSmartVolumeMapper,
)

# print("successfully finished importing vtkmodules")



"""Fit a surface to a set of points"""
# Thin Plate Spline transformations describe a nonlinear warp
# transform defined by a set of source and target landmarks.
# Any point on the mesh close to a source landmark will
# be moved to a place close to the corresponding target landmark.
# The points in between are interpolated using Bookstein's algorithm.
from vedo import Grid, Points, Arrows, show
import numpy as np
np.random.seed(1)

surf = Grid([0,0,0], res=[25,25])
ids = np.random.randint(0, surf.npoints, 10)  # pick 10 indices
pts = surf.points()[ids]

ptsource, pttarget = [], []
for pt in pts:
    pt1 = pt + [0, 0, np.random.randn() * 0.1]
    pt2 = surf.closest_point(pt1)
    ptsource.append(pt2)
    pttarget.append(pt1)

warped = surf.warp(ptsource, pttarget, mode='2d')
warped.color("b4").lc('lightblue').lw(0.1).wireframe(False)

apts = Points(pttarget, r=15, c="red5")
arrs = Arrows(ptsource, pttarget, c='k')

show(warped, apts, arrs, __doc__, axes=1, viewup="z").close()



from scipy import special
from scipy.optimize import curve_fit
from vedo import np, settings, Marker
from vedo.pyplot import plot

settings.default_font = 'Calco'
settings.remember_last_figure_format = True

xdata = [230, 234, 240, 243, 246, 249, 252]
ydata = [0,   0,    11,  62,  15,  21, 100]
tdata = [100, 31,   34, 80,   21,  21, 100]

yerrs = np.sqrt(ydata) /np.array(tdata) + 0.1
ydata = np.array(ydata) /np.array(tdata)

def func(x, a, x0):
    return (1 + special.erf(a*(x-x0))) / 2

p0 = [1/25, 240] # initial guess
popt, _ = curve_fit(func, xdata, ydata, p0)

x = np.linspace(225, 255, 50)
y = func(x, *popt)
x0, y0 = popt[1], func(popt[1], *popt)

fig = plot(
    xdata,
    ydata,
    'o',
    yerrors=yerrs,
    ylim=(-0.1,1.3),
    title="ERF(x) fit to data",
    ytitle='Embryos with visible HL',
    xtitle='Hind Limb age (h)',
    mc='blue2',
    ms=0.3,
    lwe=2,
    label='data',
)

fig += plot(x, y, lw=5, label='fit')
fig += Marker('*', s=0.05, c='r4').pos(x0,y0, 0.1)

fig.add_label(':mu', marker='*', mc="r4")
fig.add_legend("top-left", vspace=2.5)

fig.show(size=(900, 650), zoom='tight').close()





"""Cut a mesh with another mesh"""
from vedo import dataurl, settings, Plotter, Volume, Ellipsoid

settings.tiff_orientation_type = 4 # data origin is bottom-left

vol = Volume(dataurl+"embryo.tif")
iso = vol.isosurface(30, flying_edges=False).normalize()

# mesh used to cut:
emsh = Ellipsoid().scale(0.4).pos(2.8, 1.5, 1.5).wireframe()

# make a working copy and cut it with the ellipsoid
cut_embryo = iso.clone().cut_with_mesh(emsh).c("gold").bc("t")

plt = Plotter(N=2, axes=1)
plt.at(0).show(iso, emsh, viewup="z")
plt.at(1).show(cut_embryo, __doc__)
plt.interactive().close()



"""A flag-post style marker"""
from vedo import ParametricShape, precision, color_map, show

s = ParametricShape("RandomHills").cmap("coolwarm")

pts = s.clone().decimate(n=10).points()

fss = []
for p in pts:
    col = color_map(p[2], name="coolwarm", vmin=0, vmax=0.7)
    ht = precision(p[2], 3)
    fs = s.flagpost(f"Heigth:\nz={ht}m", p, c=col)
    fss.append(fs)

show(s, *fss, __doc__, bg="bb", axes=1, viewup="z")



# Available modifiers:
#  c (foreground color), bc (background color)
#  bold, blink, underLine, dim, invert, box
from vedo import printc

printc(":world: 1- Change the world by being yourself - Amy Poehler", c=1)
printc(":smile: 2- Never regret anything that made you smile - Mark Twain", c="r", bold=0)
printc(":construction: 3- Every moment is a fresh beginning - T.S Eliot", c="m", underline=1)
printc(":thumbup: 4- Die with memories, not dreams - Unknown", blink=1, bold=0)
printc(":pin: 5- When words fail, music speaks - Shakespeare")
printc(":rocket: 6- Everything you can imagine is real - Pablo Picasso", c=3)
printc(":idea: 7- Simplicity is the ultimate sophistication - Leonardo da Vinci", c=4)
printc(":rainbow: 8- Whatever you do, do it well - Walt Disney", c=3, bc=1)
printc(":target: 9- What we think, we become - Buddha", c=6, invert=1)
printc(":sparks:10- All limitations are self-imposed - Oliver Wendell Holmes", c=7, dim=1)
printc(":checked:11- If you tell the truth you don't have to remember anything - Mark Twain",
        underline=True,
        invert=True,
        c="indigo9",
)

from vedo.colors import emoji
for k in emoji.keys():
    print(emoji[k], "\t", k)



"""Create individual axes to each separate object in a scene.
Access any element to change its size and color"""
from vedo import *

settings.use_depth_peeling = True

# Create a bunch of objects
s1 = Sphere(pos=(10, 0, 0), r=1, c='r')
s2 = Sphere(pos=( 0,10, 0), r=2, c='g')
s3 = Sphere(pos=( 0, 0,10), r=3, c='b')
pt = Point([-4,-4,-4], c='k')

# Build individual axes for each object.
#  A new Assembly object is returned:
axes1 = Axes(s1, c='r')
axes2 = Axes(s2, c='g')
axes3 = Axes(s3, c='b', number_of_divisions=10)

# axes3 is an Assembly (group of Meshes).
# Unpack it and scale the 7th label getting it by its name,
# make it 5 times bigger big and fuchsia:
axes3.unpack('xNumericLabel7').scale(5).c('fuchsia')
# Print all element names in axes3:
#for m in axes3.get_meshes(): print(m.name)

# By specifiyng axes in show(), new axes are
#  created which span the whole bounding box.
#  Options are passed through a dictionary
show(pt, s1, s2, s3, axes1, axes2, axes3, __doc__,
     viewup='z',
     axes=dict(c='black', number_of_divisions=10, yzgrid=False),
).close()



import magpylib  # pip install magpylib
import numpy as np
import vedo

coil1 = magpylib.Collection()
for z in np.linspace(-8, 8, 16):
    winding = magpylib.current.Loop(
        current=100,
        diameter=10,
        position=(0,0,z),
    )
    coil1.add(winding)

#####################
volume = vedo.Volume(
    dims=(41, 41, 41),
    spacing=(2, 2, 2),
    origin=(-40, -40, -40),
)

# compute B-field and add as pointdata to volume
volume.pointdata['B'] = coil1.getB(volume.points())

# compute field lines
seeds = vedo.Disc(r1=1, r2=5.2, res=(3,12))

streamlines = vedo.StreamLines(
    volume,
    seeds,
    max_propagation=180,
    initial_step_size=0.1,
    direction="both",
)
streamlines.cmap("RdBu_r", "B").lw(5).add_scalarbar("mT")

vedo.show(streamlines, axes=1)



"""Form a surface by joining two lines"""
from vedo import *
import numpy as np

l1 = [[sin(x), cos(x), x/3] for x in np.arange(0,9, 0.1)]
l2 = [[sin(x)+0.2, cos(x) + x/15, x/3] for x in np.arange(0,9, 0.1)]

t1 = Tube(l1, c="green5", r=0.02)
t2 = Tube(l2, c="blue3",  r=0.02)

r12 = Ribbon(l1, l2, res=(200,5)).alpha(0.5)
r1  = Ribbon(l1, width=0.1).alpha(0.5).color('orange')

plt = Plotter(N=2, axes=1)
plt.at(0).show(__doc__, r12, t1, t2)
plt.at(1).show("..or along a single line", r1, t1, viewup="z")
plt.interactive().close()



"""Load an existing vtkStructuredGrid and draw
the streamlines of the velocity field"""
from vedo import *

######################## vtk
import vtk
# Read the data and specify which scalars and vectors to read.
pl3d = vtk.vtkMultiBlockPLOT3DReader()
fpath = download(dataurl+"combxyz.bin")
pl3d.SetXYZFileName(fpath)
fpath = download(dataurl+"combq.bin")
pl3d.SetQFileName(fpath)
pl3d.SetScalarFunctionNumber(100)
pl3d.SetVectorFunctionNumber(202)
pl3d.Update()
# this vtkStructuredData already has a vector field:
domain = pl3d.GetOutput().GetBlock(0)

######################## vedo
probe = Grid(s=[5,5], res=[6,6]).rotate_y(90).pos(5,0,29)

stream = StreamLines(domain, probe)

box = Mesh(domain).alpha(0.1).c('white')

show(stream, probe, box, __doc__, axes=7, bg='bb').close()




from vedo import *

ug1 = UGrid(dataurl+'ugrid.vtk')

ug2= ug1.clone().tomesh().wireframe()

cyl = Cylinder(r=3, height=7).x(3).wireframe()
ug1.cut_with_mesh(cyl)

show(ug1, ug2, cyl, axes=1).close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np

try:
    import vedo.vtkclasses as vtk
except ImportError:
    import vtkmodules.all as vtk

import vedo
from vedo.colors import printc
from vedo import utils
from vedo import shapes
from vedo.mesh import Mesh
from vedo.plotter import show

__docformat__ = "google"

__doc__ = """
Submodule for support of the [FEniCS/Dolfin](https://fenicsproject.org) library.

Example:
    ```python
    import dolfin
    from vedo import dataurl, download
    from vedo.dolfin import plot
    fname = download(dataurl+"dolfin_fine.xml")
    mesh = dolfin.Mesh(fname)
    plot(mesh)
    ```

![](https://user-images.githubusercontent.com/32848391/53026243-d2d31900-3462-11e9-9dde-518218c241b6.jpg)

.. note::
    Find many more examples in directory    
    [vedo/examples/dolfin](https://github.com/marcomusy/vedo/blob/master/examples/other/dolfin).
"""

__all__ = ["plot"]


##########################################################################
def _inputsort(obj):

    import dolfin

    u = None
    mesh = None
    if not utils.is_sequence(obj):
        obj = [obj]

    for ob in obj:
        inputtype = str(type(ob))

        # printc('inputtype is', inputtype, c=2)

        if "vedo" in inputtype:  # skip vtk objects, will be added later
            continue

        if "dolfin" in inputtype or "ufl" in inputtype:
            if "MeshFunction" in inputtype:
                mesh = ob.mesh()

                if ob.dim() > 0:
                    printc("MeshFunction of dim>0 not supported.", c="r")
                    printc('Try e.g.:  MeshFunction("size_t", mesh, 0)', c="r", italic=1)
                    printc('instead of MeshFunction("size_t", mesh, 1)', c="r", strike=1)
                else:
                    # printc(ob.dim(), mesh.num_cells(), len(mesh.coordinates()), len(ob.array()))
                    V = dolfin.FunctionSpace(mesh, "CG", 1)
                    u = dolfin.Function(V)
                    v2d = dolfin.vertex_to_dof_map(V)
                    u.vector()[v2d] = ob.array()
            elif "Function" in inputtype or "Expression" in inputtype:
                u = ob
            elif "ufl.mathfunctions" in inputtype:  # not working
                u = ob
            elif "Mesh" in inputtype:
                mesh = ob
            elif "algebra" in inputtype:
                mesh = ob.ufl_domain()
                # print('algebra', ob.ufl_domain())

    if u and not mesh and hasattr(u, "function_space"):
        V = u.function_space()
        if V:
            mesh = V.mesh()
    if u and not mesh and hasattr(u, "mesh"):
        mesh = u.mesh()

    # printc('------------------------------------')
    # printc('mesh.topology dim=', mesh.topology().dim())
    # printc('mesh.geometry dim=', mesh.geometry().dim())
    # if u: printc('u.value_rank()', u.value_rank())
    # if u and u.value_rank(): printc('u.value_dimension()', u.value_dimension(0)) # axis=0
    ##if u: printc('u.value_shape()', u.value_shape())
    return (mesh, u)


def _compute_uvalues(u, mesh):
    # the whole purpose of this function is
    # to have a scalar (or vector) for each point of the mesh

    if not u:
        return np.array([])
    #    print('u',u)

    if hasattr(u, "compute_vertex_values"):  # old dolfin, works fine
        u_values = u.compute_vertex_values(mesh)

        if u.value_rank() and u.value_dimension(0) > 1:
            l = u_values.shape[0]
            u_values = u_values.reshape(u.value_dimension(0), int(l / u.value_dimension(0))).T

    elif hasattr(u, "compute_point_values"):  # dolfinx
        u_values = u.compute_point_values()

        try:
            from dolfin import fem

            fvec = u.vector
        except RuntimeError:
            fspace = u.function_space
            try:
                fspace = fspace.collapse()
            except RuntimeError:
                return []

        fvec = fem.interpolate(u, fspace).vector

        tdim = mesh.topology.dim

        # print('fvec.getSize', fvec.getSize(), mesh.num_entities(tdim))
        if fvec.getSize() == mesh.num_entities(tdim):
            # DG0 cellwise function
            C = fvec.get_local()
            if C.dtype.type is np.complex128:
                print("Plotting real part of complex data")
                C = np.real(C)

        u_values = C

    else:
        u_values = []

    if hasattr(mesh, "coordinates"):
        coords = mesh.coordinates()
    else:
        coords = mesh.geometry.points

    if u_values.shape[0] != coords.shape[0]:
        vedo.logger.warning("mismatch in vedo.dolfin._compute_uvalues")
        u_values = np.array([u(p) for p in coords])
    return u_values


def plot(*inputobj, **options):
    """
    Plot the object(s) provided.

    Input can be any combination of: `Mesh`, `Volume`, `dolfin.Mesh`,
    `dolfin.MeshFunction`, `dolfin.Expression` or `dolfin.Function`.

    Return the current `Plotter` class instance.

    Arguments:
        mode : (str)
            one or more of the following can be combined in any order
            - `mesh`/`color`, will plot the mesh, by default colored with a scalar if available
            - `displacement` show displaced mesh by solution
            - `arrows`, mesh displacements are plotted as scaled arrows.
            - `lines`, mesh displacements are plotted as scaled lines.
            - `tensors`, to be implemented
        add : (bool)
            add the input objects without clearing the already plotted ones
        density : (float)
            show only a subset of lines or arrows [0-1]
        wire[frame] : (bool)
            visualize mesh as wireframe [False]
        c[olor] : (color)
            set mesh color [None]
        exterior : (bool)
            only show the outer surface of the mesh [False]
        alpha : (float)
            set object's transparency [1]
        lw : (int)
            line width of the mesh (set to zero to hide mesh) [0.1]
        ps :  int
            set point size of mesh vertices [None]
        z : (float)
            add a constant to z-coordinate (useful to show 2D slices as function of time)
        legend : (str)
            add a legend to the top-right of window [None]
        scalarbar : (bool)
            add a scalarbar to the window ['vertical']
        vmin : (float)
            set the minimum for the range of the scalar [None]
        vmax : (float)
            set the maximum for the range of the scalar [None]
        scale : (float)
            add a scaling factor to arrows and lines sizes [1]
        cmap : (str)
            choose a color map for scalars
        shading : (str)
            mesh shading ['flat', 'phong']
        text : (str)
            add a gray text comment to the top-left of the window [None]
        isolines : (dict)
            dictionary of isolines properties
            - n, (int) - add this number of isolines to the mesh
            - c, - isoline color
            - lw, (float) - isoline width
            - z, (float) - add to the isoline z coordinate to make them more visible
        streamlines : (dict)
            dictionary of streamlines properties
            - probes, (list, None) - custom list of points to use as seeds
            - tol, (float) - tolerance to reduce the number of seed points used in mesh
            - lw, (float) - line width of the streamline
            - direction, (str) - direction of integration ('forward', 'backward' or 'both')
            - max_propagation, (float) - max propagation of the streamline
            - scalar_range, (list) - scalar range of coloring
        warpZfactor : (float)
            elevate z-axis by scalar value (useful for 2D geometries)
        warpYfactor : (float)
            elevate z-axis by scalar value (useful for 1D geometries)
        scaleMeshFactors : (list)
            rescale mesh by these factors [1,1,1]
        new : (bool)
            spawn a new instance of Plotter class, pops up a new window
        at : (int)
            renderer number to plot to
        shape : (list)
            subdvide window in (n,m) rows and columns
        N : (int)
            automatically subdvide window in N renderers
        pos : (list)
            (x,y) coordinates of the window position on screen
        size : (list)
            window size (x,y)
        title : (str)
            window title
        bg : (color)
            background color name of window
        bg2 : (color)
            second background color name to create a color gradient
        style : (int)
            choose a predefined style [0-4]
            - 0, `vedo`, style (blackboard background, rainbow color map)
            - 1, `matplotlib`, style (white background, viridis color map)
            - 2, `paraview`, style
            - 3, `meshlab`, style
            - 4, `bw`, black and white style.
        axes : (int)
            Axes type number.
            Axes type-1 can be fully customized by passing a dictionary `axes=dict()`.
            - 0,  no axes,
            - 1,  draw customizable grid axes (see below).
            - 2,  show cartesian axes from (0,0,0)
            - 3,  show positive range of cartesian axes from (0,0,0)
            - 4,  show a triad at bottom left
            - 5,  show a cube at bottom left
            - 6,  mark the corners of the bounding box
            - 7,  draw a simple ruler at the bottom of the window
            - 8,  show the `vtkCubeAxesActor` object,
            - 9,  show the bounding box outLine,
            - 10, show three circles representing the maximum bounding box,
            - 11, show a large grid on the x-y plane (use with zoom=8)
            - 12, show polar axes.
        infinity : (bool)
            if True fugue point is set at infinity (no perspective effects)
        sharecam : (bool)
            if False each renderer will have an independent vtkCamera
        interactive : (bool)
            if True will stop after show() to allow interaction w/ window
        offscreen : (bool)
            if True will not show the rendering window
        zoom : (float)
            camera zooming factor
        viewup : (list), str
            camera view-up direction ['x','y','z', or a vector direction]
        azimuth : (float)
            add azimuth rotation of the scene, in degrees
        elevation : (float)
            add elevation rotation of the scene, in degrees
        roll : (float)
            add roll-type rotation of the scene, in degrees
        camera : (dict)
            Camera parameters can further be specified with a dictionary
            assigned to the `camera` keyword:
            (E.g. `show(camera={'pos':(1,2,3), 'thickness':1000,})`)
            - `pos`, `(list)`,
                the position of the camera in world coordinates
            - `focal_point`, `(list)`,
                the focal point of the camera in world coordinates
            - `viewup`, `(list)`,
                the view up direction for the camera
            - `distance`, `(float)`,
                set the focal point to the specified distance from the camera position.
            - `clipping_range`, `(float)`,
                distance of the near and far clipping planes along the direction of projection.
            - `parallel_scale`, `(float)`,
                scaling used for a parallel projection, i.e. the height of the viewport
                in world-coordinate distances. The default is 1. Note that the "scale" parameter works as
                an "inverse scale", larger numbers produce smaller images.
                This method has no effect in perspective projection mode.
            - `thickness`, `(float)`,
                set the distance between clipping planes. This method adjusts the far clipping
                plane to be set a distance 'thickness' beyond the near clipping plane.
            - `view_angle`, `(float)`,
                the camera view angle, which is the angular height of the camera view
                measured in degrees. The default angle is 30 degrees.
                This method has no effect in parallel projection mode.
                The formula for setting the angle up for perfect perspective viewing is:
                angle = 2*atan((h/2)/d) where h is the height of the RenderWindow
                (measured by holding a ruler up to your screen) and d is the distance
                from your eyes to the screen.
    """
    if len(inputobj) == 0:
        vedo.plotter_instance.interactive()
        return None

    if "numpy" in str(type(inputobj[0])):
        from vedo.pyplot import plot as pyplot_plot

        return pyplot_plot(*inputobj, **options)

    mesh, u = _inputsort(inputobj)

    mode = options.pop("mode", "mesh")
    ttime = options.pop("z", None)

    add = options.pop("add", False)

    wire = options.pop("wireframe", None)

    c = options.pop("c", None)
    color = options.pop("color", None)
    if color is not None:
        c = color

    lc = options.pop("lc", None)

    alpha = options.pop("alpha", 1)
    lw = options.pop("lw", 0.5)
    ps = options.pop("ps", None)
    legend = options.pop("legend", None)
    scbar = options.pop("scalarbar", "v")
    vmin = options.pop("vmin", None)
    vmax = options.pop("vmax", None)
    cmap = options.pop("cmap", None)
    scale = options.pop("scale", 1)
    scaleMeshFactors = options.pop("scaleMeshFactors", [1, 1, 1])
    shading = options.pop("shading", "phong")
    text = options.pop("text", None)
    style = options.pop("style", "vtk")
    isolns = options.pop("isolines", {})
    streamlines = options.pop("streamlines", {})
    warpZfactor = options.pop("warpZfactor", None)
    warpYfactor = options.pop("warpYfactor", None)
    lighting = options.pop("lighting", None)
    exterior = options.pop("exterior", False)
    returnActorsNoShow = options.pop("returnActorsNoShow", False)
    at = options.pop("at", 0)

    # refresh axes titles for axes type = 8 (vtkCubeAxesActor)
    xtitle = options.pop("xtitle", "x")
    ytitle = options.pop("ytitle", "y")
    ztitle = options.pop("ztitle", "z")
    if vedo.plotter_instance:
        if xtitle != "x":
            aet = vedo.plotter_instance.axes_instances
            if len(aet) > at and isinstance(aet[at], vtk.vtkCubeAxesActor):
                aet[at].SetXTitle(xtitle)
        if ytitle != "y":
            aet = vedo.plotter_instance.axes_instances
            if len(aet) > at and isinstance(aet[at], vtk.vtkCubeAxesActor):
                aet[at].SetYTitle(ytitle)
        if ztitle != "z":
            aet = vedo.plotter_instance.axes_instances
            if len(aet) > at and isinstance(aet[at], vtk.vtkCubeAxesActor):
                aet[at].SetZTitle(ztitle)

    # change some default to emulate standard behaviours
    if style in (0, "vtk"):
        axes = options.pop("axes", None)
        if axes is None:
            options["axes"] = {"xygrid": False, "yzgrid": False, "zxgrid": False}
        else:
            options["axes"] = axes  # put back
        if cmap is None:
            cmap = "rainbow"
    elif style in (1, "matplotlib"):
        bg = options.pop("bg", None)
        if bg is None:
            options["bg"] = "white"
        else:
            options["bg"] = bg
        axes = options.pop("axes", None)
        if axes is None:
            options["axes"] = {"xygrid": False, "yzgrid": False, "zxgrid": False}
        else:
            options["axes"] = axes  # put back
        if cmap is None:
            cmap = "viridis"
    elif style in (2, "paraview"):
        bg = options.pop("bg", None)
        if bg is None:
            options["bg"] = (82, 87, 110)
        else:
            options["bg"] = bg
        if cmap is None:
            cmap = "coolwarm"
    elif style in (3, "meshlab"):
        bg = options.pop("bg", None)
        if bg is None:
            options["bg"] = (8, 8, 16)
            options["bg2"] = (117, 117, 234)
        else:
            options["bg"] = bg
        axes = options.pop("axes", None)
        if axes is None:
            options["axes"] = 10
        else:
            options["axes"] = axes  # put back
        if cmap is None:
            cmap = "afmhot"
    elif style in (4, "bw"):
        bg = options.pop("bg", None)
        if bg is None:
            options["bg"] = (217, 255, 238)
        else:
            options["bg"] = bg
        axes = options.pop("axes", None)
        if axes is None:
            options["axes"] = {"xygrid": False, "yzgrid": False, "zxgrid": False}
        else:
            options["axes"] = axes  # put back
        if cmap is None:
            cmap = "binary"

    #################################################################
    actors = []
    if vedo.plotter_instance:
        if add:
            actors = vedo.plotter_instance.actors

    if mesh and ("mesh" in mode or "color" in mode or "displace" in mode):

        actor = MeshActor(u, mesh, exterior=exterior)

        actor.wireframe(wire)
        actor.scale(scaleMeshFactors)
        if lighting:
            actor.lighting(lighting)
        if ttime:
            actor.z(ttime)
        if legend:
            actor.legend(legend)
        if c:
            actor.color(c)
        if lc:
            actor.linecolor(lc)
        if alpha:
            alpha = min(alpha, 1)
            actor.alpha(alpha * alpha)
        if lw:
            actor.linewidth(lw)
            if wire and alpha:
                lw1 = min(lw, 1)
                actor.alpha(alpha * lw1)
        if ps:
            actor.point_size(ps)
        if shading:
            if shading == "phong":
                actor.phong()
            elif shading == "flat":
                actor.flat()
            elif shading[0] == "g":
                actor.gouraud()

        if "displace" in mode:
            actor.move(u)

        if cmap and (actor.u_values is not None) and len(actor.u_values)>0 and c is None:
            if actor.u_values.ndim > 1:
                actor.cmap(cmap, utils.mag(actor.u_values), vmin=vmin, vmax=vmax)
            else:
                actor.cmap(cmap, actor.u_values, vmin=vmin, vmax=vmax)

        if warpYfactor:
            scals = actor.pointdata[0]
            if len(scals) > 0:
                pts_act = actor.points()
                pts_act[:, 1] = scals * warpYfactor * scaleMeshFactors[1]
        if warpZfactor:
            scals = actor.pointdata[0]
            if len(scals) > 0:
                pts_act = actor.points()
                pts_act[:, 2] = scals * warpZfactor * scaleMeshFactors[2]
        if warpYfactor or warpZfactor:
            actor.points(pts_act)
            if vmin is not None and vmax is not None:
                actor.mapper().SetScalarRange(vmin, vmax)

        if scbar and c is None:
            if "3d" in scbar:
                actor.add_scalarbar3d()
            elif "h" in scbar:
                actor.add_scalarbar(horizontal=True)
            else:
                actor.add_scalarbar(horizontal=False)

        if len(isolns) > 0:
            ison = isolns.pop("n", 10)
            isocol = isolns.pop("c", "black")
            isoalpha = isolns.pop("alpha", 1)
            isolw = isolns.pop("lw", 1)

            isos = actor.isolines(n=ison).color(isocol).lw(isolw).alpha(isoalpha)

            isoz = isolns.pop("z", None)
            if isoz is not None:  # kind of hack to make isolines visible on flat meshes
                d = isoz
            else:
                d = actor.diagonal_size() / 400
            isos.z(actor.z() + d)
            actors.append(isos)

        actors.append(actor)

    #################################################################
    if "streamline" in mode:
        mode = mode.replace("streamline", "")
        str_act = MeshStreamLines(u, **streamlines)
        actors.append(str_act)

    #################################################################
    if "arrow" in mode or "line" in mode:
        if "arrow" in mode:
            arrs = MeshArrows(u, scale=scale)
        else:
            arrs = MeshLines(u, scale=scale)

        if arrs:
            if legend and "mesh" not in mode:
                arrs.legend(legend)
            if c:
                arrs.color(c)
                arrs.color(c)
            if alpha:
                arrs.alpha(alpha)
            actors.append(arrs)

    #################################################################
    if "tensor" in mode:
        pass  # todo

    #################################################################
    for ob in inputobj:
        inputtype = str(type(ob))
        if "vedo" in inputtype:
            actors.append(ob)

    if text:
        # textact = Text2D(text, font=font)
        actors.append(text)

    if "at" in options and "interactive" not in options:
        if vedo.plotter_instance:
            N = vedo.plotter_instance.shape[0] * vedo.plotter_instance.shape[1]
            if options["at"] == N - 1:
                options["interactive"] = True

    # if vedo.plotter_instance:
    #     for a2 in vedo.collectable_actors:
    #         if isinstance(a2, vtk.vtkCornerAnnotation):
    #             if 0 in a2.rendered_at: # remove old message
    #                 vedo.plotter_instance.remove(a2)
    #                 break

    if len(actors) == 0:
        print('Warning: no objects to show, check mode in plot(mode="...")')

    if returnActorsNoShow:
        return actors

    return show(actors, **options)


###################################################################################
class MeshActor(Mesh):
    """MeshActor for dolfin support."""

    def __init__(self, *inputobj, **options):
        """MeshActor, a `vedo.Mesh` derived object for dolfin support."""

        c = options.pop("c", None)
        alpha = options.pop("alpha", 1)
        exterior = options.pop("exterior", False)
        compute_normals = options.pop("compute_normals", False)

        mesh, u = _inputsort(inputobj)
        if not mesh:
            return

        if exterior:
            import dolfin

            meshc = dolfin.BoundaryMesh(mesh, "exterior")
        else:
            meshc = mesh

        if hasattr(mesh, "coordinates"):
            coords = mesh.coordinates()
        else:
            coords = mesh.geometry.points

        cells = meshc.cells()

        if cells.shape[1] == 4:
            # something wrong in this as it cannot reproduce the tet cell..
            # from vedo.tetmesh import _buildtetugrid
            # cells[:,[2, 0]] = cells[:,[0, 2]]
            # cells[:,[1, 0]] = cells[:,[0, 1]]
            # cells[:,[0, 1, 2, 3]] = cells[:,[0, 2, 1, 3]]
            # cells[:,[0, 1, 2, 3]] = cells[:,[0, 2, 1, 3]]
            # cells[:,[0, 1, 2, 3]] = cells[:,  [0, 1, 3, 2]]
            # cells[:,[0, 1, 2, 3]] = cells[:,[1, 0, 2, 3]]
            # cells[:,[0, 1, 2, 3]] = cells[:,[2, 0, 1, 3]]
            # cells[:,[0, 1, 2, 3]] = cells[:,[2, 0, 1, 3]]
            # print(cells[0])
            # print(coords[cells[0]])
            # poly = utils.geometry(_buildtetugrid(coords, cells))
            # poly = utils.geometry(vedo.TetMesh([coords, cells]).inputdata())

            poly = vtk.vtkPolyData()

            source_points = vtk.vtkPoints()
            source_points.SetData(utils.numpy2vtk(coords, dtype=np.float32))
            poly.SetPoints(source_points)

            source_polygons = vtk.vtkCellArray()
            for f in cells:
                # do not use vtkTetra() because it fails
                # with dolfin faces orientation
                ele0 = vtk.vtkTriangle()
                ele1 = vtk.vtkTriangle()
                ele2 = vtk.vtkTriangle()
                ele3 = vtk.vtkTriangle()

                f0, f1, f2, f3 = f
                pid0 = ele0.GetPointIds()
                pid1 = ele1.GetPointIds()
                pid2 = ele2.GetPointIds()
                pid3 = ele3.GetPointIds()

                pid0.SetId(0, f0)
                pid0.SetId(1, f1)
                pid0.SetId(2, f2)

                pid1.SetId(0, f0)
                pid1.SetId(1, f1)
                pid1.SetId(2, f3)

                pid2.SetId(0, f1)
                pid2.SetId(1, f2)
                pid2.SetId(2, f3)

                pid3.SetId(0, f2)
                pid3.SetId(1, f3)
                pid3.SetId(2, f0)

                source_polygons.InsertNextCell(ele0)
                source_polygons.InsertNextCell(ele1)
                source_polygons.InsertNextCell(ele2)
                source_polygons.InsertNextCell(ele3)

            poly.SetPolys(source_polygons)

        else:
            poly = utils.buildPolyData(coords, cells)

        Mesh.__init__(self, poly, c=c, alpha=alpha)
        if compute_normals:
            self.compute_normals()

        self.mesh = mesh  # holds a dolfin Mesh obj
        self.u = u  # holds a dolfin function_data
        # holds the actual values of u on the mesh
        self.u_values = _compute_uvalues(u, mesh)

    def move(self, u=None, deltas=None):
        """Move mesh according to solution `u` or from calculated vertex displacements `deltas`."""
        if u is None:
            u = self.u
        if deltas is None:
            if self.u_values is not None:
                deltas = self.u_values
            else:
                deltas = _compute_uvalues(u, self.mesh)
                self.u_values = deltas

        if hasattr(self.mesh, "coordinates"):
            coords = self.mesh.coordinates()
        else:
            coords = self.mesh.geometry.points

        if coords.shape != deltas.shape:
            vedo.logger.error(
                f"Try to move mesh with wrong solution type shape {coords.shape} vs {deltas.shape}"
            )
            vedo.logger.error("Mesh is not moved. Try mode='color' in plot().")
            return

        movedpts = coords + deltas
        if movedpts.shape[1] == 2:  # 2d
            movedpts = np.c_[movedpts, np.zeros(movedpts.shape[0])]
        self.polydata(False).GetPoints().SetData(utils.numpy2vtk(movedpts, dtype=np.float32))
        self.polydata(False).GetPoints().Modified()


def MeshPoints(*inputobj, **options):
    """Build a point object of type `Mesh` for a list of points."""
    r = options.pop("r", 5)
    c = options.pop("c", "gray")
    alpha = options.pop("alpha", 1)

    mesh, u = _inputsort(inputobj)
    if not mesh:
        return None

    if hasattr(mesh, "coordinates"):
        plist = mesh.coordinates()
    else:
        plist = mesh.geometry.points

    u_values = _compute_uvalues(u, mesh)

    if len(plist[0]) == 2:  # coords are 2d.. not good..
        plist = np.insert(plist, 2, 0, axis=1)  # make it 3d
    if len(plist[0]) == 1:  # coords are 1d.. not good..
        plist = np.insert(plist, 1, 0, axis=1)  # make it 3d
        plist = np.insert(plist, 2, 0, axis=1)

    actor = shapes.Points(plist, r=r, c=c, alpha=alpha)

    actor.mesh = mesh
    if u:
        actor.u = u
        if len(u_values.shape) == 2:
            if u_values.shape[1] in [2, 3]:  # u_values is 2D or 3D
                actor.u_values = u_values
                dispsizes = utils.mag(u_values)
        else:  # u_values is 1D
            dispsizes = u_values
        actor.pointdata["u_values"] = dispsizes
        actor.pointdata.select("u_values")
    return actor


def MeshLines(*inputobj, **options):
    """
    Build the line segments between two lists of points `start_points` and `end_points`.
    `start_points` can be also passed in the form `[[point1, point2], ...]`.

    A dolfin `Mesh` that was deformed/modified by a function can be
    passed together as inputs.

    Use `scale` to apply a rescaling factor to the length
    """
    scale = options.pop("scale", 1)
    lw = options.pop("lw", 1)
    c = options.pop("c", "grey")
    alpha = options.pop("alpha", 1)

    mesh, u = _inputsort(inputobj)
    if not mesh:
        return None

    if hasattr(mesh, "coordinates"):
        start_points = mesh.coordinates()
    else:
        start_points = mesh.geometry.points

    u_values = _compute_uvalues(u, mesh)
    if not utils.is_sequence(u_values[0]):
        vedo.logger.error("cannot show Lines for 1D scalar values")
        raise RuntimeError()

    end_points = start_points + u_values
    if u_values.shape[1] == 2:  # u_values is 2D
        u_values = np.insert(u_values, 2, 0, axis=1)  # make it 3d
        start_points = np.insert(start_points, 2, 0, axis=1)  # make it 3d
        end_points = np.insert(end_points, 2, 0, axis=1)  # make it 3d

    actor = shapes.Lines(start_points, end_points, scale=scale, lw=lw, c=c, alpha=alpha)

    actor.mesh = mesh
    actor.u = u
    actor.u_values = u_values
    return actor


def MeshArrows(*inputobj, **options):
    """Build arrows representing displacements."""
    s = options.pop("s", None)
    c = options.pop("c", "gray")
    scale = options.pop("scale", 1)
    alpha = options.pop("alpha", 1)
    res = options.pop("res", 12)

    mesh, u = _inputsort(inputobj)
    if not mesh:
        return None

    if hasattr(mesh, "coordinates"):
        start_points = mesh.coordinates()
    else:
        start_points = mesh.geometry.points

    u_values = _compute_uvalues(u, mesh)
    if not utils.is_sequence(u_values[0]):
        vedo.logger.error("cannot show Arrows for 1D scalar values")
        raise RuntimeError()

    end_points = start_points + u_values * scale
    if u_values.shape[1] == 2:  # u_values is 2D
        u_values = np.insert(u_values, 2, 0, axis=1)  # make it 3d
        start_points = np.insert(start_points, 2, 0, axis=1)  # make it 3d
        end_points = np.insert(end_points, 2, 0, axis=1)  # make it 3d

    actor = shapes.Arrows(start_points, end_points, s=s, alpha=alpha, res=res)
    actor.color(c)
    actor.mesh = mesh
    actor.u = u
    actor.u_values = u_values
    return actor


def MeshStreamLines(*inputobj, **options):
    """Build a streamplot."""
    from vedo.shapes import StreamLines

    print("Building streamlines...")

    tol = options.pop("tol", 0.02)
    lw = options.pop("lw", 2)
    direction = options.pop("direction", "forward")
    max_propagation = options.pop("max_propagation", None)
    scalar_range = options.pop("scalar_range", None)
    probes = options.pop("probes", None)

    tubes = options.pop("tubes", {})  # todo
    maxRadiusFactor = options.pop("maxRadiusFactor", 1)
    varyRadius = options.pop("varyRadius", 1)

    mesh, u = _inputsort(inputobj)
    if not mesh:
        return None

    u_values = _compute_uvalues(u, mesh)
    if not utils.is_sequence(u_values[0]):
        vedo.logger.error("cannot show Arrows for 1D scalar values")
        raise RuntimeError()
    if u_values.shape[1] == 2:  # u_values is 2D
        u_values = np.insert(u_values, 2, 0, axis=1)  # make it 3d

    meshact = MeshActor(u)
    meshact.pointdata["u_values"] = u_values
    meshact.pointdata.select("u_values")

    if utils.is_sequence(probes):
        pass  # it's already it
    elif tol:
        print("decimating mesh points to use them as seeds...")
        probes = meshact.clone().subsample(tol).points()
    else:
        probes = meshact.points()
    if len(probes) > 500:
        printc("Probing domain with n =", len(probes), "points")
        printc(" ..this may take time (or choose a larger tol value)")

    if lw:
        tubes = {}
    else:
        tubes["varyRadius"] = varyRadius
        tubes["maxRadiusFactor"] = maxRadiusFactor

    str_lns = StreamLines(
        meshact,
        probes,
        direction=direction,
        max_propagation=max_propagation,
        tubes=tubes,
        scalar_range=scalar_range,
        active_vectors="u_values",
    )

    if lw:
        str_lns.lw(lw)

    return str_lns



"""Render meshes into inset windows
(which can be dragged)"""
from vedo import *

plt = Plotter(bg2='bisque', size=(1000,800), interactive=False)

e = Volume(dataurl+"embryo.tif").isosurface()
e.normalize().shift(-2,-1.5,-2).c("gold")

plt.show(e, __doc__, viewup='z')

# make clone copies of the embryo surface and cut them:
e1 = e.clone().cut_with_plane(normal=[0,1,0]).c("green4")
e2 = e.clone().cut_with_plane(normal=[1,0,0]).c("red5")

# add 2 draggable inset windows:
plt.add_inset(e1, pos=(0.9,0.8))
plt.add_inset(e2, pos=(0.9,0.5))

# customised axes can also be inserted:
ax = Axes(
    xrange=(0,1), yrange=(0,1), zrange=(0,1),
    xtitle='front', ytitle='left', ztitle='head',
    yzgrid=False, xtitle_size=0.15, ytitle_size=0.15, ztitle_size=0.15,
    xlabel_size=0, ylabel_size=0, zlabel_size=0, tip_size=0.05,
    axes_linewidth=2, xline_color='dr', yline_color='dg', zline_color='db',
    xtitle_offset=0.05, ytitle_offset=0.05, ztitle_offset=0.05,
)

ex = e.clone().scale(0.25).pos(0,0.1,0.1).alpha(0.1).lighting('off')
plt.add_inset(ax, ex, pos=(0.1,0.1), size=0.15, draggable=False)
plt.interactive().close()



"""Customizing Axes.
Cartesian planes can be displaced
from their lower-range default position"""
from vedo import Sphere, Axes, precision, show

sph = Sphere().scale([4, 3, 2]).shift(5, 6, 7).c("green2", 0.1).wireframe()

axs = Axes(
    sph,  # build axes for object sph
    xtitle="x axis",
    ytitle="y axis",
    ztitle="z axis",
    htitle="An ellipsoid at " + precision(sph.center_of_mass(), 2),
    htitle_font=1,
    htitle_color="red3",
    zxgrid=True,
    xyframe_line=2,
    yzframe_line=2,
    zxframe_line=2,
    xyframe_color="red3",
    yzframe_color="green3",
    zxframe_color="blue3",
    xyshift=0.2,  # move xy plane 20% along z
    yzshift=0.2,  # move yz plane 20% along x
    zxshift=0.2,  # move zx plane 20% along y
)

show(sph, axs, __doc__).close()



"""
Show a cube for each available texture name.
Any jpg file can be used as texture.
"""
from vedo import dataurl, show, Cube

textures_path = dataurl+'textures/'

print(__doc__)
print('example textures:', textures_path)

cubes = []
cubes.append(Cube().texture(textures_path+'leather.jpg'))
cubes.append(Cube().texture(textures_path+'paper2.jpg'))
cubes.append(Cube().texture(textures_path+'wood1.jpg'))
cubes.append(Cube().texture(textures_path+'wood2.jpg'))

show(cubes, N=4, bg2='lightblue').close()



"""Visualize a n:dotm numpy matrix"""
from vedo.pyplot import matrix
from vedo import show
import numpy as np

n, m = (6, 5)
M = np.eye(n, m)/2 + np.random.randn(n, m)*0.1
# print(M)

mat = matrix(
    M,
    cmap='Reds',
    title='My numpy Matrix',
    xtitle='Genes of group A',
    ytitle='Genes of group B',
    xlabels=[f'hox{i}' for i in range(m)],
    ylabels=[f'bmp{i}' for i in range(n)],
    scale=0.15,  # size of bin labels; set it to 0 to remove labels
    lw=2,        # separator line width
)

show(mat, __doc__, bg='k7', zoom=1.2).close()



"""
Example of drawing objects on different windows
and/or subwindows within the same window.
We split the main window in many subwindows and draw
somethingon specific windows numbers.
Then open an independent window and draw a shape on it.
"""
from vedo import *

##########################################################################
# this is one instance of the class Plotter with 5 raws and 5 columns
plt1 = Plotter(shape=(5,5), axes=0)

# set a different background color for a specific subwindow (the last one)
plt1.background([0.8, 0.9, 0.9], at=24)

# load the meshes and give them a name
a = Mesh(dataurl+"shuttle.obj")
b = Mesh(dataurl+"cessna.vtk").c("red")
c = Mesh(dataurl+"porsche.ply")

# show a Text2D in each renderer
for i in range(25):
    plt1.at(i).show(f"renderer\nnr.{i}")

plt1.at( 6).show(a)
plt1.at(23).show(b)
plt1.at(24).show(c)


##########################################################################
# declare a second independent instance of the class Plotter
# shape can also be given as a string, e.g.:
# shape="2/6" means 2 renderers above and 6 below
# shape="3|1" means 3 renderers on the left and one on the right

s = Mesh(dataurl+'mug.ply')

# Set the position of the horizontal of vertical splitting [0,1]:
#settings.window_splitting_position = 0.5

plt2 = Plotter(pos=(500, 250), shape='2/6')

for i in range(len(plt2.renderers)):
    s2 = s.clone(deep=False).color(i)
    plt2.at(i).show(s2, f'renderer #{i}')

printc(__doc__)
plt2.interactive()
plt2.close()
plt1.close()



#!/usr/bin/env python
#
from trame.app import get_server
from trame.ui.vuetify import SinglePageLayout
from trame.widgets import vtk, vuetify

from vedo import Volume, Axes, Plotter, dataurl

vol = Volume(dataurl+"embryo.slc")

plt = Plotter(bg='Wheat')
plt += [vol, Axes(vol)]
plt += vol.isosurface().shift(300,0,0)

# ------------------------------------------------------------
# Web Application setup
# ------------------------------------------------------------
server = get_server()
ctrl = server.controller

with SinglePageLayout(server) as layout:
    layout.title.set_text("Hello trame")

    with layout.content:
        with vuetify.VContainer(
            fluid=True,
            classes="pa-0 fill-height",
        ):
            plt.reset_camera()
            view = vtk.VtkRemoteView(plt.window)
            # view = vtk.VtkLocalView(plt.window)

server.start()



"""Lock an object orientation
to constantly face the scene camera"""
from vedo import *

sp = Sphere().wireframe()
pts = sp.points()

tx1 = Text3D("Fixed Text", pts[10], s=0.07, depth=0.1, c="lb")
tx2 = Text3D("Follower Text", pts[144], s=0.07, c="lg").follow_camera()

fp = sp.flagpole("The\nNorth Pole", c='k6', rounded=True).scale(0.4).follow_camera()

show(sp, tx1, tx2, fp, __doc__, bg='bb', axes=1).close()



"""Metrics of quality for
the cells of a triangular mesh
(zoom to see cell label values)"""
from vedo import dataurl, Mesh, show
from vedo.pyplot import histogram

mesh = Mesh(dataurl + "panther.stl").compute_normals().linewidth(0.1).flat()

# generate a numpy array for mesh quality
mesh.compute_quality(metric=6)
mesh.cmap("RdYlBu", on="cells")

hist = histogram(mesh.celldata["Quality"], xtitle="mesh quality", ac="w")
# make it smaller and position it, use_bounds makes the cam
# ignore the object when resetting the 3d qscene
hist.scale(0.6).pos(40, -53, 0).use_bounds(False)

# add a scalar bar for the active scalars
mesh.add_scalarbar3d(c="w", title="triangle quality by min(:alpha_i )")

# create numeric labels of active scalar on top of cells
labs = mesh.labels(on="cells", precision=3, scale=0.4, font="Quikhand", c="black")

cam = dict(
    pos=(59.8, -191, 78.9),
    focal_point=(27.9, -2.94, 3.33),
    viewup=(-0.0170, 0.370, 0.929),
    distance=205,
    clipping_range=(87.8, 355),
)

show(mesh, labs, hist, __doc__, bg="bb", camera=cam, axes=11).close()



"""Double pendulum from ODE integration"""
# Copyright (c) 2018, N. Rougier, https://github.com/rougier/pendulum
# http://www.physics.usyd.edu.au/~wheat/dpend_html/solve_dpend.c
# Adapted for vedo by M. Musy, 2021
from scipy import integrate
from vedo import *

G  = 9.81   # acceleration due to gravity, in m/s^2
L1 = 1.0    # length of pendulum 1 in m
L2 = 1.0    # length of pendulum 2 in m
M1 = 1.0    # mass of pendulum 1 in kg
M2 = 1.0    # mass of pendulum 2 in kg
th1= 120    # initial angles (degrees)
th2= -20
w1 = 0      # initial angular velocities (degrees per second)
w2 = 0
dt = 0.015


def derivs(state, t):
    dydx = np.zeros_like(state)
    dydx[0] = state[1]
    a = state[2] - state[0]
    sina, cosa = sin(a), cos(a)
    den1 = (M1 + M2)*L1 - M2*L1*cosa*cosa
    dydx[1] = (M2*L1*state[1]*state[1]*sina*cosa +
               M2*G*sin(state[2])*cosa +
               M2*L2*state[3]*state[3]*sina -
               (M1+M2)*G*sin(state[0]) )/den1
    dydx[2] = state[3]
    den2 = (L2/L1)*den1
    dydx[3] = (-M2*L2*state[3]*state[3]*sina*cosa +
               (M1+M2)*G*sin(state[0])*cosa -
               (M1+M2)*L1*state[1]*state[1]*sina -
               (M1+M2)*G*sin(state[2]) )/den2
    return dydx

t = np.arange(0.0, 10.0, dt)
state = np.radians([th1, w1, th2, w2])
y = integrate.odeint(derivs, state, t)

P1 = np.dstack([L1*sin(y[:,0]), -L1*cos(y[:,0])]).squeeze()
P2 = P1 + np.dstack([L2*sin(y[:,2]), -L2*cos(y[:,2])]).squeeze()

plt = Plotter(interactive=False, size=(900,700),)
ax = Axes(xrange=(-2,2), yrange=(-2,1), htitle=__doc__)

for i in progressbar(len(t)):
    j = max(i- 5,0)
    k = max(i-10,0)
    l1 = Line([[0,0], P1[i], P2[i]]).lw(7).c("blue2", 1.0)
    l2 = Line([[0,0], P1[j], P2[j]]).lw(6).c("blue2", 0.4)
    l3 = Line([[0,0], P1[k], P2[k]]).lw(5).c("blue2", 0.2)
    plt.clear().show(l1, l2, l3, ax, zoom=1.4)

plt.interactive().close()



# Transform a picture into a mesh
from vedo import Picture, dataurl, show
import numpy as np

pic = Picture(dataurl+"images/dog.jpg").smooth(5)
msh = pic.tomesh()  # make a quad-mesh out of it

# build a scalar array with intensities
rgb = msh.pointdata["RGBA"]
intensity = np.sum(rgb, axis=1)
intensityz = np.zeros_like(rgb)
intensityz[:,2] = intensity / 10

# set the new vertex points
pts = msh.points() + intensityz
msh.points(pts)

# more cosmetics
msh.triangulate().smooth()
msh.lighting("default").linewidth(0)
msh.cmap("bone", "RGBA").add_scalarbar()

msht = pic.clone().threshold(100).linewidth(0)

show([[pic, "A normal jpg image.."],
      [msh, "..becomes a polygonal Mesh"],
      [msht, "Thresholding also generates a Mesh"]
     ], N=3, axes=1, zoom=1.1, elevation=-20, bg='black').close()



import glob
import os
from deprecated import deprecated

import numpy as np

try:
    import vedo.vtkclasses as vtk
except ImportError:
    import vtkmodules.all as vtk

import vedo
from vedo import utils
from vedo.base import Base3DProp
from vedo.base import BaseGrid
from vedo.mesh import Mesh


__docformat__ = "google"

__doc__ = """
Work with volumetric datasets (voxel data).

![](https://vedo.embl.es/images/volumetric/slicePlane2.png)
"""

__all__ = ["BaseVolume", "Volume", "VolumeSlice"]


##########################################################################
class BaseVolume:
    """
    Base class. Do not instantiate.
    """

    def __init__(self):
        """Base class. Do not instantiate."""

        self._data = None
        self._mapper = None
        self.transform = None
        self.pipeline = None

    def _repr_html_(self):
        """
        HTML representation of the Volume object for Jupyter Notebooks.

        Returns:
            HTML text with the image and some properties.
        """
        import io
        import base64
        from PIL import Image

        library_name = "vedo.volume.Volume"
        help_url = "https://vedo.embl.es/docs/vedo/volume.html"

        arr = self.thumbnail(azimuth=0, elevation=-60, zoom=1.4, axes=True)

        im = Image.fromarray(arr)
        buffered = io.BytesIO()
        im.save(buffered, format="PNG", quality=100)
        encoded = base64.b64encode(buffered.getvalue()).decode("utf-8")
        url = "data:image/png;base64," + encoded
        image = f"<img src='{url}'></img>"

        # statisitics
        bounds = "<br/>".join(
            [
                utils.precision(min_x, 4) + " ... " + utils.precision(max_x, 4)
                for min_x, max_x in zip(self.bounds()[::2], self.bounds()[1::2])
            ]
        )

        help_text = ""
        if self.name:
            help_text += f"<b> {self.name}: &nbsp&nbsp</b>"
        help_text += '<b><a href="' + help_url + '" target="_blank">' + library_name + "</a></b>"
        if self.filename:
            dots = ""
            if len(self.filename) > 30:
                dots = "..."
            help_text += f"<br/><code><i>({dots}{self.filename[-30:]})</i></code>"

        pdata = ""
        if self._data.GetPointData().GetScalars():
            if self._data.GetPointData().GetScalars().GetName():
                name = self._data.GetPointData().GetScalars().GetName()
                pdata = "<tr><td><b> point data array </b></td><td>" + name + "</td></tr>"

        cdata = ""
        if self._data.GetCellData().GetScalars():
            if self._data.GetCellData().GetScalars().GetName():
                name = self._data.GetCellData().GetScalars().GetName()
                cdata = "<tr><td><b> voxel data array </b></td><td>" + name + "</td></tr>"

        img = self.GetMapper().GetInput()

        allt = [
            "<table>",
            "<tr>",
            "<td>",
            image,
            "</td>",
            "<td style='text-align: center; vertical-align: center;'><br/>",
            help_text,
            "<table>",
            "<tr><td><b> bounds </b> <br/> (x/y/z) </td><td>" + str(bounds) + "</td></tr>",
            "<tr><td><b> dimensions </b></td><td>" + str(img.GetDimensions()) + "</td></tr>",
            "<tr><td><b> voxel spacing </b></td><td>"
            + utils.precision(img.GetSpacing(), 3)
            + "</td></tr>",
            "<tr><td><b> in memory size </b></td><td>"
            + str(int(img.GetActualMemorySize() / 1024))
            + "MB</td></tr>",
            pdata,
            cdata,
            "<tr><td><b> scalar range </b></td><td>"
            + utils.precision(img.GetScalarRange(), 4)
            + "</td></tr>",
            "</table>",
            "</table>",
        ]
        return "\n".join(allt)

    def _update(self, img):
        self._data = img
        self._data.GetPointData().Modified()
        self._mapper.SetInputData(img)
        self._mapper.Modified()
        self._mapper.Update()
        return self

    def clone(self):
        """Return a clone copy of the Volume."""
        newimg = vtk.vtkImageData()
        newimg.CopyStructure(self._data)
        newimg.CopyAttributes(self._data)

        newvol = Volume(newimg)
        prop = vtk.vtkVolumeProperty()
        prop.DeepCopy(self.GetProperty())
        newvol.SetProperty(prop)
        newvol.SetOrigin(self.GetOrigin())
        newvol.SetScale(self.GetScale())
        newvol.SetOrientation(self.GetOrientation())
        newvol.SetPosition(self.GetPosition())
        newvol.pipeline = utils.OperationNode("clone", parents=[self], c="#bbd0ff", shape="diamond")
        return newvol

    def imagedata(self):
        """Return the underlying `vtkImagaData` object."""
        return self._data

    def tonumpy(self):
        """
        Get read-write access to voxels of a Volume object as a numpy array.

        When you set values in the output image, you don't want numpy to reallocate the array
        but instead set values in the existing array, so use the [:] operator.

        Example:
            `arr[:] = arr*2 + 15`

        If the array is modified add a call to:
        `volume.imagedata().GetPointData().GetScalars().Modified()`
        when all your modifications are completed.
        """
        narray_shape = tuple(reversed(self._data.GetDimensions()))

        scals = self._data.GetPointData().GetScalars()
        comps = scals.GetNumberOfComponents()
        if comps == 1:
            narray = utils.vtk2numpy(scals).reshape(narray_shape)
            narray = np.transpose(narray, axes=[2, 1, 0])
        else:
            narray = utils.vtk2numpy(scals).reshape(*narray_shape, comps)
            narray = np.transpose(narray, axes=[2, 1, 0, 3])

        # narray = utils.vtk2numpy(self._data.GetPointData().GetScalars()).reshape(narray_shape)
        # narray = np.transpose(narray, axes=[2, 1, 0])

        return narray

    def dimensions(self):
        """Return the nr. of voxels in the 3 dimensions."""
        return np.array(self._data.GetDimensions())

    @deprecated(reason=vedo.colors.red + "Please use scalar_range()" + vedo.colors.reset)
    def scalarRange(self):
        "Deprecated. Please use `scalar_range()`."
        return self.scalar_range()

    def scalar_range(self):
        """Return the range of the scalar values."""
        return np.array(self._data.GetScalarRange())

    def spacing(self, s=None):
        """Set/get the voxels size in the 3 dimensions."""
        if s is not None:
            self._data.SetSpacing(s)
            return self
        return np.array(self._data.GetSpacing())

    # def pos(self, p=None): # conflicts with API of base.x()
    #     """Same effect as calling `origin()`."""
    #     return self.origin(p)

    def origin(self, s=None):
        """Set/get the origin of the volumetric dataset."""
        ### supersedes base.origin()
        ### DIFFERENT from base.origin()!
        if s is not None:
            self._data.SetOrigin(s)
            return self
        return np.array(self._data.GetOrigin())

    def center(self, p=None):
        """Set/get the center of the volumetric dataset."""
        if p is not None:
            self._data.SetCenter(p)
            return self
        return np.array(self._data.GetCenter())

    def permute_axes(self, x, y, z):
        """
        Reorder the axes of the Volume by specifying
        the input axes which are supposed to become the new X, Y, and Z.
        """
        imp = vtk.vtkImagePermute()
        imp.SetFilteredAxes(x, y, z)
        imp.SetInputData(self.imagedata())
        imp.Update()
        self._update(imp.GetOutput())
        self.pipeline = utils.OperationNode(
            f"permute_axes\n{(x,y,z)}", parents=[self], c="#4cc9f0"
        )
        return self

    def resample(self, new_spacing, interpolation=1):
        """
        Resamples a `Volume` to be larger or smaller.

        This method modifies the spacing of the input.
        Linear interpolation is used to resample the data.

        Arguments:
            new_spacing : (list)
                a list of 3 new spacings for the 3 axes
            interpolation : (int)
                0=nearest_neighbor, 1=linear, 2=cubic
        """
        rsp = vtk.vtkImageResample()
        oldsp = self.spacing()
        for i in range(3):
            if oldsp[i] != new_spacing[i]:
                rsp.SetAxisOutputSpacing(i, new_spacing[i])
        rsp.InterpolateOn()
        rsp.SetInterpolationMode(interpolation)
        rsp.OptimizationOn()
        rsp.Update()
        self._update(rsp.GetOutput())
        self.pipeline = utils.OperationNode(
            f"resample\n{tuple(new_spacing)}", parents=[self], c="#4cc9f0"
        )
        return self

    def interpolation(self, itype):
        """
        Set interpolation type.

        0=nearest neighbour, 1=linear
        """
        self.property.SetInterpolationType(itype)
        return self

    def threshold(self, above=None, below=None, replace=None, replace_value=None):
        """
        Binary or continuous volume thresholding.
        Find the voxels that contain a value above/below the input values
        and replace them with a new value (default is 0).
        """
        th = vtk.vtkImageThreshold()
        th.SetInputData(self.imagedata())

        # sanity checks
        if above is not None and below is not None:
            if above == below:
                return self
            if above > below:
                vedo.logger.warning("in volume.threshold(), above > below, skip.")
                return self

        ## cases
        if below is not None and above is not None:
            th.ThresholdBetween(above, below)

        elif above is not None:
            th.ThresholdByUpper(above)

        elif below is not None:
            th.ThresholdByLower(below)

        ##
        if replace is not None:
            th.SetReplaceIn(True)
            th.SetInValue(replace)
        else:
            th.SetReplaceIn(False)

        if replace_value is not None:
            th.SetReplaceOut(True)
            th.SetOutValue(replace_value)
        else:
            th.SetReplaceOut(False)

        th.Update()
        self._update(th.GetOutput())
        self.pipeline = utils.OperationNode("threshold", parents=[self], c="#4cc9f0")
        return self

    def crop(self, left=None, right=None, back=None, front=None, bottom=None, top=None, VOI=()):
        """
        Crop a `Volume` object.

        Arguments:
            left : (float)
                fraction to crop from the left plane (negative x)
            right : (float)
                fraction to crop from the right plane (positive x)
            back : (float)
                fraction to crop from the back plane (negative y)
            front : (float)
                fraction to crop from the front plane (positive y)
            bottom : (float)
                fraction to crop from the bottom plane (negative z)
            top : (float)
                fraction to crop from the top plane (positive z)
            VOI : (list)
                extract Volume Of Interest expressed in voxel numbers

        Example:
            `vol.crop(VOI=(xmin, xmax, ymin, ymax, zmin, zmax)) # all integers nrs`
        """
        extractVOI = vtk.vtkExtractVOI()
        extractVOI.SetInputData(self.imagedata())

        if VOI:
            extractVOI.SetVOI(VOI)
        else:
            d = self.imagedata().GetDimensions()
            bx0, bx1, by0, by1, bz0, bz1 = 0, d[0]-1, 0, d[1]-1, 0, d[2]-1
            if left is not None:   bx0 = int((d[0]-1)*left)
            if right is not None:  bx1 = int((d[0]-1)*(1-right))
            if back is not None:   by0 = int((d[1]-1)*back)
            if front is not None:  by1 = int((d[1]-1)*(1-front))
            if bottom is not None: bz0 = int((d[2]-1)*bottom)
            if top is not None:    bz1 = int((d[2]-1)*(1-top))
            extractVOI.SetVOI(bx0, bx1, by0, by1, bz0, bz1)
        extractVOI.Update()
        self._update(extractVOI.GetOutput())

        self.pipeline = utils.OperationNode(
            "crop", parents=[self], c="#4cc9f0", comment=f"dims={tuple(self.dimensions())}"
        )
        return self

    def append(self, volumes, axis="z", preserve_extents=False):
        """
        Take the components from multiple inputs and merges them into one output.
        Except for the append axis, all inputs must have the same extent.
        All inputs must have the same number of scalar components.
        The output has the same origin and spacing as the first input.
        The origin and spacing of all other inputs are ignored.
        All inputs must have the same scalar type.

        Arguments:
            axis : (int, str)
                axis expanded to hold the multiple images
            preserve_extents : (bool)
                if True, the extent of the inputs is used to place
                the image in the output. The whole extent of the output is the union of the input
                whole extents. Any portion of the output not covered by the inputs is set to zero.
                The origin and spacing is taken from the first input.

        Example:
            ```python
            from vedo import Volume, dataurl
            vol = Volume(dataurl+'embryo.tif')
            vol.append(vol, axis='x').show().close()
            ```
            ![](https://vedo.embl.es/images/feats/volume_append.png)
        """
        ima = vtk.vtkImageAppend()
        ima.SetInputData(self.imagedata())
        if not utils.is_sequence(volumes):
            volumes = [volumes]
        for volume in volumes:
            if isinstance(volume, vtk.vtkImageData):
                ima.AddInputData(volume)
            else:
                ima.AddInputData(volume.imagedata())
        ima.SetPreserveExtents(preserve_extents)
        if axis == "x":
            axis = 0
        elif axis == "y":
            axis = 1
        elif axis == "z":
            axis = 2
        ima.SetAppendAxis(axis)
        ima.Update()
        self._update(ima.GetOutput())

        self.pipeline = utils.OperationNode(
            "append",
            parents=[self, *volumes],
            c="#4cc9f0",
            comment=f"dims={tuple(self.dimensions())}",
        )
        return self

    def pad(self, voxels=10, value=0):
        """
        Add the specified number of voxels at the `Volume` borders.
        Voxels can be a list formatted as `[nx0, nx1, ny0, ny1, nz0, nz1]`.

        Arguments:
            voxels : (int, list)
                number of voxels to be added (or a list of length 4)
            value : (int)
                intensity value (gray-scale color) of the padding
        
        Example:
            ```python
            from vedo import Volume, dataurl, show
            iso = Volume(dataurl+'embryo.tif').isosurface()
            vol = iso.binarize(spacing=(100, 100, 100)).pad(10)
            vol.dilate([15,15,15])
            show(iso, vol.isosurface(), N=2, axes=1)
            ```
            ![](https://vedo.embl.es/images/volumetric/volume_pad.png)
        """
        x0, x1, y0, y1, z0, z1 = self._data.GetExtent()
        pf = vtk.vtkImageConstantPad()
        pf.SetInputData(self._data)
        pf.SetConstant(value)
        if utils.is_sequence(voxels):
            pf.SetOutputWholeExtent(
                x0 - voxels[0], x1 + voxels[1], 
                y0 - voxels[2], y1 + voxels[3], 
                z0 - voxels[4], z1 + voxels[5], 
            )
        else:
            pf.SetOutputWholeExtent(
                x0 - voxels, x1 + voxels, 
                y0 - voxels, y1 + voxels, 
                z0 - voxels, z1 + voxels,
            )
        pf.Update()
        self._update(pf.GetOutput())
        self.pipeline = utils.OperationNode(
            "pad", comment=f"{voxels} voxels", parents=[self], c="#f28482"
        )
        return self

    def resize(self, *newdims):
        """Increase or reduce the number of voxels of a Volume with interpolation."""
        old_dims = np.array(self.imagedata().GetDimensions())
        old_spac = np.array(self.imagedata().GetSpacing())
        rsz = vtk.vtkImageResize()
        rsz.SetResizeMethodToOutputDimensions()
        rsz.SetInputData(self.imagedata())
        rsz.SetOutputDimensions(newdims)
        rsz.Update()
        self._data = rsz.GetOutput()
        new_spac = old_spac * old_dims / newdims  # keep aspect ratio
        self._data.SetSpacing(new_spac)
        self._update(self._data)
        self.pipeline = utils.OperationNode(
            "resize", parents=[self], c="#4cc9f0", comment=f"dims={tuple(self.dimensions())}"
        )
        return self

    def normalize(self):
        """Normalize that scalar components for each point."""
        norm = vtk.vtkImageNormalize()
        norm.SetInputData(self.imagedata())
        norm.Update()
        self._update(norm.GetOutput())
        self.pipeline = utils.OperationNode("normalize", parents=[self], c="#4cc9f0")
        return self

    def mirror(self, axis="x"):
        """
        Mirror flip along one of the cartesian axes.
        """
        img = self.imagedata()

        ff = vtk.vtkImageFlip()
        ff.SetInputData(img)
        if axis.lower() == "x":
            ff.SetFilteredAxis(0)
        elif axis.lower() == "y":
            ff.SetFilteredAxis(1)
        elif axis.lower() == "z":
            ff.SetFilteredAxis(2)
        else:
            vedo.logger.error("mirror must be set to either x, y, z or n")
            raise RuntimeError()
        ff.Update()
        self._update(ff.GetOutput())
        self.pipeline = utils.OperationNode(f"mirror {axis}", parents=[self], c="#4cc9f0")
        return self

    def operation(self, operation, volume2=None):
        """
        Perform operations with `Volume` objects.
        Keyword `volume2` can be a constant float.

        Possible operations are:
        ```
        +, -, /, 1/x, sin, cos, exp, log,
        abs, **2, sqrt, min, max, atan, atan2, median,
        mag, dot, gradient, divergence, laplacian.
        ```

        Examples:
            - [volume_operations.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/volume_operations.py)
        """
        op = operation.lower()
        image1 = self._data

        mf = None
        if op in ["median"]:
            mf = vtk.vtkImageMedian3D()
            mf.SetInputData(image1)
        elif op in ["mag"]:
            mf = vtk.vtkImageMagnitude()
            mf.SetInputData(image1)
        elif op in ["dot", "dotproduct"]:
            mf = vtk.vtkImageDotProduct()
            mf.SetInput1Data(image1)
            mf.SetInput2Data(volume2.imagedata())
        elif op in ["grad", "gradient"]:
            mf = vtk.vtkImageGradient()
            mf.SetDimensionality(3)
            mf.SetInputData(image1)
        elif op in ["div", "divergence"]:
            mf = vtk.vtkImageDivergence()
            mf.SetInputData(image1)
        elif op in ["laplacian"]:
            mf = vtk.vtkImageLaplacian()
            mf.SetDimensionality(3)
            mf.SetInputData(image1)

        if mf is not None:
            mf.Update()
            vol = Volume(mf.GetOutput())
            vol.pipeline = utils.OperationNode(
                f"operation\n{op}", parents=[self], c="#4cc9f0", shape="cylinder"
            )
            return vol  ###########################

        mat = vtk.vtkImageMathematics()
        mat.SetInput1Data(image1)

        K = None

        if utils.is_number(volume2):
            K = volume2
            mat.SetConstantK(K)
            mat.SetConstantC(K)

        elif volume2 is not None:  # assume image2 is a constant value
            mat.SetInput2Data(volume2.imagedata())

        # ###########################
        if op in ["+", "add", "plus"]:
            if K:
                mat.SetOperationToAddConstant()
            else:
                mat.SetOperationToAdd()

        elif op in ["-", "subtract", "minus"]:
            if K:
                mat.SetConstantC(-float(K))
                mat.SetOperationToAddConstant()
            else:
                mat.SetOperationToSubtract()

        elif op in ["*", "multiply", "times"]:
            if K:
                mat.SetOperationToMultiplyByK()
            else:
                mat.SetOperationToMultiply()

        elif op in ["/", "divide"]:
            if K:
                mat.SetConstantK(1.0 / K)
                mat.SetOperationToMultiplyByK()
            else:
                mat.SetOperationToDivide()

        elif op in ["1/x", "invert"]:
            mat.SetOperationToInvert()
        elif op in ["sin"]:
            mat.SetOperationToSin()
        elif op in ["cos"]:
            mat.SetOperationToCos()
        elif op in ["exp"]:
            mat.SetOperationToExp()
        elif op in ["log"]:
            mat.SetOperationToLog()
        elif op in ["abs"]:
            mat.SetOperationToAbsoluteValue()
        elif op in ["**2", "square"]:
            mat.SetOperationToSquare()
        elif op in ["sqrt", "sqr"]:
            mat.SetOperationToSquareRoot()
        elif op in ["min"]:
            mat.SetOperationToMin()
        elif op in ["max"]:
            mat.SetOperationToMax()
        elif op in ["atan"]:
            mat.SetOperationToATAN()
        elif op in ["atan2"]:
            mat.SetOperationToATAN2()
        else:
            vedo.logger.error(f"unknown operation {operation}")
            raise RuntimeError()
        mat.Update()

        self._update(mat.GetOutput())

        self.pipeline = utils.OperationNode(
            f"operation\n{op}", parents=[self, volume2], shape="cylinder", c="#4cc9f0"
        )
        return self

    def frequency_pass_filter(self, low_cutoff=None, high_cutoff=None, order=1):
        """
        Low-pass and high-pass filtering become trivial in the frequency domain.
        A portion of the pixels/voxels are simply masked or attenuated.
        This function applies a high pass Butterworth filter that attenuates the
        frequency domain image.

        The gradual attenuation of the filter is important.
        A simple high-pass filter would simply mask a set of pixels in the frequency domain,
        but the abrupt transition would cause a ringing effect in the spatial domain.

        Arguments:
            low_cutoff : (list)
                the cutoff frequencies for x, y and z
            high_cutoff : (list)
                the cutoff frequencies for x, y and z
            order : (int)
                order determines sharpness of the cutoff curve
        """
        # https://lorensen.github.io/VTKExamples/site/Cxx/ImageProcessing/IdealHighPass
        fft = vtk.vtkImageFFT()
        fft.SetInputData(self._data)
        fft.Update()
        out = fft.GetOutput()

        if high_cutoff:
            blp = vtk.vtkImageButterworthLowPass()
            blp.SetInputData(out)
            blp.SetCutOff(high_cutoff)
            blp.SetOrder(order)
            blp.Update()
            out = blp.GetOutput()

        if low_cutoff:
            bhp = vtk.vtkImageButterworthHighPass()
            bhp.SetInputData(out)
            bhp.SetCutOff(low_cutoff)
            bhp.SetOrder(order)
            bhp.Update()
            out = bhp.GetOutput()

        rfft = vtk.vtkImageRFFT()
        rfft.SetInputData(out)
        rfft.Update()

        ecomp = vtk.vtkImageExtractComponents()
        ecomp.SetInputData(rfft.GetOutput())
        ecomp.SetComponents(0)
        ecomp.Update()
        self._update(ecomp.GetOutput())
        self.pipeline = utils.OperationNode("frequency_pass_filter", parents=[self], c="#4cc9f0")
        return self

    def smooth_gaussian(self, sigma=(2, 2, 2), radius=None):
        """
        Performs a convolution of the input Volume with a gaussian.

        Arguments:
            sigma : (float, list)
                standard deviation(s) in voxel units.
                A list can be given to smooth in the three direction differently.
            radius : (float, list)
                radius factor(s) determine how far out the gaussian
                kernel will go before being clamped to zero. A list can be given too.
        """
        gsf = vtk.vtkImageGaussianSmooth()
        gsf.SetDimensionality(3)
        gsf.SetInputData(self.imagedata())
        if utils.is_sequence(sigma):
            gsf.SetStandardDeviations(sigma)
        else:
            gsf.SetStandardDeviation(sigma)
        if radius is not None:
            if utils.is_sequence(radius):
                gsf.SetRadiusFactors(radius)
            else:
                gsf.SetRadiusFactor(radius)
        gsf.Update()
        self._update(gsf.GetOutput())
        self.pipeline = utils.OperationNode("smooth_gaussian", parents=[self], c="#4cc9f0")
        return self

    def smooth_median(self, neighbours=(2, 2, 2)):
        """
        Median filter that replaces each pixel with the median value
        from a rectangular neighborhood around that pixel.
        """
        imgm = vtk.vtkImageMedian3D()
        imgm.SetInputData(self.imagedata())
        if utils.is_sequence(neighbours):
            imgm.SetKernelSize(neighbours[0], neighbours[1], neighbours[2])
        else:
            imgm.SetKernelSize(neighbours, neighbours, neighbours)
        imgm.Update()
        self._update(imgm.GetOutput())
        self.pipeline = utils.OperationNode("smooth_median", parents=[self], c="#4cc9f0")
        return self

    def erode(self, neighbours=(2, 2, 2)):
        """
        Replace a voxel with the minimum over an ellipsoidal neighborhood of voxels.
        If `neighbours` of an axis is 1, no processing is done on that axis.

        Examples:
            - [erode_dilate.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/erode_dilate.py)

                ![](https://vedo.embl.es/images/volumetric/erode_dilate.png)
        """
        ver = vtk.vtkImageContinuousErode3D()
        ver.SetInputData(self._data)
        ver.SetKernelSize(neighbours[0], neighbours[1], neighbours[2])
        ver.Update()
        self._update(ver.GetOutput())
        self.pipeline = utils.OperationNode("erode", parents=[self], c="#4cc9f0")
        return self

    def dilate(self, neighbours=(2, 2, 2)):
        """
        Replace a voxel with the maximum over an ellipsoidal neighborhood of voxels.
        If `neighbours` of an axis is 1, no processing is done on that axis.

        Check also `erode()` and `pad()`.

        Examples:
            - [erode_dilate.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/erode_dilate.py)
        """
        ver = vtk.vtkImageContinuousDilate3D()
        ver.SetInputData(self._data)
        ver.SetKernelSize(neighbours[0], neighbours[1], neighbours[2])
        ver.Update()
        self._update(ver.GetOutput())
        self.pipeline = utils.OperationNode("dilate", parents=[self], c="#4cc9f0")
        return self

    def magnitude(self):
        """Colapses components with magnitude function."""
        imgm = vtk.vtkImageMagnitude()
        imgm.SetInputData(self.imagedata())
        imgm.Update()
        self._update(imgm.GetOutput())
        self.pipeline = utils.OperationNode("magnitude", parents=[self], c="#4cc9f0")
        return self

    def topoints(self):
        """
        Extract all image voxels as points.
        This function takes an input `Volume` and creates an `Mesh`
        that contains the points and the point attributes.

        Examples:
            - [vol2points.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/vol2points.py)
        """
        v2p = vtk.vtkImageToPoints()
        v2p.SetInputData(self.imagedata())
        v2p.Update()
        mpts = vedo.Points(v2p.GetOutput())
        mpts.pipeline = utils.OperationNode("topoints", parents=[self], c="#4cc9f0:#e9c46a")
        return mpts

    def euclidean_distance(self, anisotropy=False, max_distance=None):
        """
        Implementation of the Euclidean DT (Distance Transform) using Saito's algorithm.
        The distance map produced contains the square of the Euclidean distance values.
        The algorithm has a O(n^(D+1)) complexity over n x n x...x n images in D dimensions.

        Check out also: https://en.wikipedia.org/wiki/Distance_transform

        Arguments:
            anisotropy : bool
                used to define whether Spacing should be used in the
                computation of the distances.
            max_distance : bool
                any distance bigger than max_distance will not be
                computed but set to this specified value instead.

        Examples:
            - [euclidian_dist.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/euclidian_dist.py)
        """
        euv = vtk.vtkImageEuclideanDistance()
        euv.SetInputData(self._data)
        euv.SetConsiderAnisotropy(anisotropy)
        if max_distance is not None:
            euv.InitializeOn()
            euv.SetMaximumDistance(max_distance)
        euv.SetAlgorithmToSaito()
        euv.Update()
        vol = Volume(euv.GetOutput())
        vol.pipeline = utils.OperationNode("euclidean_distance", parents=[self], c="#4cc9f0")
        return vol

    def correlation_with(self, vol2, dim=2):
        """
        Find the correlation between two volumetric data sets.
        Keyword `dim` determines whether the correlation will be 3D, 2D or 1D.
        The default is a 2D Correlation.

        The output size will match the size of the first input.
        The second input is considered the correlation kernel.
        """
        imc = vtk.vtkImageCorrelation()
        imc.SetInput1Data(self._data)
        imc.SetInput2Data(vol2.imagedata())
        imc.SetDimensionality(dim)
        imc.Update()
        vol = Volume(imc.GetOutput())

        vol.pipeline = utils.OperationNode("correlation_with", parents=[self, vol2], c="#4cc9f0")
        return vol

    def scale_voxels(self, scale=1):
        """Scale the voxel content by factor `scale`."""
        rsl = vtk.vtkImageReslice()
        rsl.SetInputData(self.imagedata())
        rsl.SetScalarScale(scale)
        rsl.Update()
        self._update(rsl.GetOutput())
        self.pipeline = utils.OperationNode(
            f"scale_voxels\nscale={scale}", parents=[self], c="#4cc9f0"
        )
        return self


##########################################################################
class Volume(BaseVolume, BaseGrid, vtk.vtkVolume):
    """
    Class to describe dataset that are defined on "voxels":
    the 3D equivalent of 2D pixels.
    """

    def __init__(
        self,
        inputobj=None,
        c="RdBu_r",
        alpha=(0.0, 0.0, 0.2, 0.4, 0.8, 1.0),
        alpha_gradient=None,
        alpha_unit=1,
        mode=0,
        spacing=None,
        dims=None,
        origin=None,
        mapper="smart",
    ):
        """
        This class can be initialized with a numpy object, a `vtkImageData`
        or a list of 2D bmp files.

        Arguments:
            c : (list, str)
                sets colors along the scalar range, or a matplotlib color map name
            alphas : (float, list)
                sets transparencies along the scalar range
            alpha_unit : (float)
                low values make composite rendering look brighter and denser
            origin : (list)
                set volume origin coordinates
            spacing : (list)
                voxel dimensions in x, y and z.
            dims : (list)
                specify the dimensions of the volume.
            mapper : (str)
                either 'gpu', 'opengl_gpu', 'fixed' or 'smart'
            mode : (int)
                define the volumetric rendering style:
                    - 0, composite rendering
                    - 1, maximum projection
                    - 2, minimum projection
                    - 3, average projection
                    - 4, additive mode

                <br>The default mode is "composite" where the scalar values are sampled through
                the volume and composited in a front-to-back scheme through alpha blending.
                The final color and opacity is determined using the color and opacity transfer
                functions specified in alpha keyword.

                Maximum and minimum intensity blend modes use the maximum and minimum
                scalar values, respectively, along the sampling ray.
                The final color and opacity is determined by passing the resultant value
                through the color and opacity transfer functions.

                Additive blend mode accumulates scalar values by passing each value
                through the opacity transfer function and then adding up the product
                of the value and its opacity. In other words, the scalar values are scaled
                using the opacity transfer function and summed to derive the final color.
                Note that the resulting image is always grayscale i.e. aggregated values
                are not passed through the color transfer function.
                This is because the final value is a derived value and not a real data value
                along the sampling ray.

                Average intensity blend mode works similar to the additive blend mode where
                the scalar values are multiplied by opacity calculated from the opacity
                transfer function and then added.
                The additional step here is to divide the sum by the number of samples
                taken through the volume.
                As is the case with the additive intensity projection, the final image will
                always be grayscale i.e. the aggregated values are not passed through the
                color transfer function.

        Example:
            ```python
            from vedo import Volume
            vol = Volume("path/to/mydata/rec*.bmp")
            vol.show()
            ```

        Examples:
            - [numpy2volume1.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/numpy2volume1.py)

                ![](https://vedo.embl.es/images/volumetric/numpy2volume1.png)

            - [read_volume2.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/read_volume2.py)

                ![](https://vedo.embl.es/images/volumetric/read_volume2.png)

        .. note::
            if a `list` of values is used for `alphas` this is interpreted
            as a transfer function along the range of the scalar.
        """
        vtk.vtkVolume.__init__(self)
        BaseGrid.__init__(self)
        BaseVolume.__init__(self)
        # super().__init__()

        ###################
        if isinstance(inputobj, str):

            if "https://" in inputobj:
                inputobj = vedo.file_io.download(inputobj, verbose=False)  # fpath
            elif os.path.isfile(inputobj):
                pass
            else:
                inputobj = sorted(glob.glob(inputobj))

        ###################
        if "gpu" in mapper:
            self._mapper = vtk.vtkGPUVolumeRayCastMapper()
        elif "opengl_gpu" in mapper:
            self._mapper = vtk.vtkOpenGLGPUVolumeRayCastMapper()
        elif "smart" in mapper:
            self._mapper = vtk.vtkSmartVolumeMapper()
        elif "fixed" in mapper:
            self._mapper = vtk.vtkFixedPointVolumeRayCastMapper()
        elif isinstance(mapper, vtk.vtkMapper):
            self._mapper = mapper
        else:
            print("Error unknown mapper type", [mapper])
            raise RuntimeError()
        self.SetMapper(self._mapper)

        ###################
        inputtype = str(type(inputobj))

        # print('Volume inputtype', inputtype, c='b')

        if inputobj is None:
            img = vtk.vtkImageData()

        elif utils.is_sequence(inputobj):

            if isinstance(inputobj[0], str) and ".bmp" in inputobj[0].lower():
                # scan sequence of BMP files
                ima = vtk.vtkImageAppend()
                ima.SetAppendAxis(2)
                pb = utils.ProgressBar(0, len(inputobj))
                for i in pb.range():
                    f = inputobj[i]
                    if "_rec_spr.bmp" in f:
                        continue
                    picr = vtk.vtkBMPReader()
                    picr.SetFileName(f)
                    picr.Update()
                    mgf = vtk.vtkImageMagnitude()
                    mgf.SetInputData(picr.GetOutput())
                    mgf.Update()
                    ima.AddInputData(mgf.GetOutput())
                    pb.print("loading...")
                ima.Update()
                img = ima.GetOutput()

            else:

                if len(inputobj.shape) == 1:
                    varr = utils.numpy2vtk(inputobj)
                else:
                    varr = utils.numpy2vtk(inputobj.ravel(order="F"))
                varr.SetName("input_scalars")

                img = vtk.vtkImageData()
                if dims is not None:
                    img.SetDimensions(dims[2], dims[1], dims[0])
                else:
                    if len(inputobj.shape) == 1:
                        vedo.logger.error("must set dimensions (dims keyword) in Volume")
                        raise RuntimeError()
                    img.SetDimensions(inputobj.shape)
                img.GetPointData().AddArray(varr)
                img.GetPointData().SetActiveScalars(varr.GetName())

        elif "ImageData" in inputtype:
            img = inputobj

        elif isinstance(inputobj, Volume):
            img = inputobj.inputdata()

        elif "UniformGrid" in inputtype:
            img = inputobj

        elif hasattr(inputobj, "GetOutput"):  # passing vtk object, try extract imagdedata
            if hasattr(inputobj, "Update"):
                inputobj.Update()
            img = inputobj.GetOutput()

        elif isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = vedo.file_io.download(inputobj, verbose=False)
            img = vedo.file_io.loadImageData(inputobj)

        else:
            vedo.logger.error(f"cannot understand input type {inputtype}")
            return

        if dims is not None:
            img.SetDimensions(dims)

        if origin is not None:
            img.SetOrigin(origin)  ### DIFFERENT from volume.origin()!

        if spacing is not None:
            img.SetSpacing(spacing)

        self._data = img
        self._mapper.SetInputData(img)

        if img.GetPointData().GetScalars():
            if img.GetPointData().GetScalars().GetNumberOfComponents() == 1:
                self.mode(mode).color(c).alpha(alpha).alpha_gradient(alpha_gradient)
                self.GetProperty().SetShade(True)
                self.GetProperty().SetInterpolationType(1)
                self.GetProperty().SetScalarOpacityUnitDistance(alpha_unit)

        # remember stuff:
        self._mode = mode
        self._color = c
        self._alpha = alpha
        self._alpha_grad = alpha_gradient
        self._alpha_unit = alpha_unit

        self.pipeline = utils.OperationNode(
            "Volume", comment=f"dims={tuple(self.dimensions())}", c="#4cc9f0"
        )
        #######################################################################

    def _update(self, data):
        self._data = data
        self._data.GetPointData().Modified()
        self._mapper.SetInputData(data)
        self._mapper.Modified()
        self._mapper.Update()
        return self

    def mode(self, mode=None):
        """
        Define the volumetric rendering mode following this:
            - 0, composite rendering
            - 1, maximum projection rendering
            - 2, minimum projection rendering
            - 3, average projection rendering
            - 4, additive mode

        The default mode is "composite" where the scalar values are sampled through
        the volume and composited in a front-to-back scheme through alpha blending.
        The final color and opacity is determined using the color and opacity transfer
        functions specified in alpha keyword.

        Maximum and minimum intensity blend modes use the maximum and minimum
        scalar values, respectively, along the sampling ray.
        The final color and opacity is determined by passing the resultant value
        through the color and opacity transfer functions.

        Additive blend mode accumulates scalar values by passing each value
        through the opacity transfer function and then adding up the product
        of the value and its opacity. In other words, the scalar values are scaled
        using the opacity transfer function and summed to derive the final color.
        Note that the resulting image is always grayscale i.e. aggregated values
        are not passed through the color transfer function.
        This is because the final value is a derived value and not a real data value
        along the sampling ray.

        Average intensity blend mode works similar to the additive blend mode where
        the scalar values are multiplied by opacity calculated from the opacity
        transfer function and then added.
        The additional step here is to divide the sum by the number of samples
        taken through the volume.
        As is the case with the additive intensity projection, the final image will
        always be grayscale i.e. the aggregated values are not passed through the
        color transfer function.
        """
        if mode is None:
            return self._mapper.GetBlendMode()

        if isinstance(mode, str):
            if "comp" in mode:
                mode = 0
            elif "proj" in mode:
                if "max" in mode:
                    mode = 1
                elif "min" in mode:
                    mode = 2
                elif "ave" in mode:
                    mode = 3
                else:
                    vedo.logger.warning(f"unknown mode {mode}")
                    mode = 0
            elif "add" in mode:
                mode = 4
            else:
                vedo.logger.warning(f"unknown mode {mode}")
                mode = 0

        self._mapper.SetBlendMode(mode)
        self._mode = mode
        return self

    def shade(self, status=None):
        """
        Set/Get the shading of a Volume.
        Shading can be further controlled with `volume.lighting()` method.

        If shading is turned on, the mapper may perform shading calculations.
        In some cases shading does not apply
        (for example, in maximum intensity projection mode).
        """
        if status is None:
            return self.GetProperty().GetShade()
        self.GetProperty().SetShade(status)
        return self

    def cmap(self, c, alpha=None, vmin=None, vmax=None):
        """Same as `color()`.

        Arguments:
            alpha : (list)
                use a list to specify transparencies along the scalar range
            vmin : (float)
                force the min of the scalar range to be this value
            vmax : (float)
                force the max of the scalar range to be this value
        """
        return self.color(c, alpha, vmin, vmax)

    def jittering(self, status=None):
        """
        If `True`, each ray traversal direction will be perturbed slightly
        using a noise-texture to get rid of wood-grain effects.
        """
        if hasattr(self._mapper, "SetUseJittering"):  # tetmesh doesnt have it
            if status is None:
                return self._mapper.GetUseJittering()
            self._mapper.SetUseJittering(status)
        return self

    def mask(self, data):
        """
        Mask a volume visualization with a binary value.
        Needs to specify keyword mapper='gpu'.

        Example:
        ```python
            from vedo import np, Volume, show
            data_matrix = np.zeros([75, 75, 75], dtype=np.uint8)
            # all voxels have value zero except:
            data_matrix[0:35,   0:35,  0:35] = 1
            data_matrix[35:55, 35:55, 35:55] = 2
            data_matrix[55:74, 55:74, 55:74] = 3
            vol = Volume(data_matrix, c=['white','b','g','r'], mapper='gpu')
            data_mask = np.zeros_like(data_matrix)
            data_mask[10:65, 10:45, 20:75] = 1
            vol.mask(data_mask)
            show(vol, axes=1).close()
        ```
        See also:
            `volume.hide_voxels()`
        """
        mask = Volume(data.astype(np.uint8))
        try:
            self.mapper().SetMaskTypeToBinary()
            self.mapper().SetMaskInput(mask.inputdata())
        except AttributeError:
            vedo.logger.error("volume.mask() must create the volume with Volume(..., mapper='gpu')")
        return self

    def hide_voxels(self, ids):
        """
        Hide voxels (cells) from visualization.

        Example:
            ```python
            from vedo import *
            embryo = Volume(dataurl+'embryo.tif')
            embryo.hide_voxels(list(range(10000)))
            show(embryo, axes=1).close()
            ```

        See also:
            `volume.mask()`
        """
        ghost_mask = np.zeros(self.ncells, dtype=np.uint8)
        ghost_mask[ids] = vtk.vtkDataSetAttributes.HIDDENCELL
        name = vtk.vtkDataSetAttributes.GhostArrayName()
        garr = utils.numpy2vtk(ghost_mask, name=name, dtype=np.uint8)
        self._data.GetCellData().AddArray(garr)
        self._data.GetCellData().Modified()
        return self

    def alpha_gradient(self, alpha_grad, vmin=None, vmax=None):
        """
        Assign a set of tranparencies to a volume's gradient
        along the range of the scalar value.
        A single constant value can also be assigned.
        The gradient function is used to decrease the opacity
        in the "flat" regions of the volume while maintaining the opacity
        at the boundaries between material types.  The gradient is measured
        as the amount by which the intensity changes over unit distance.

        The format for alpha_grad is the same as for method `volume.alpha()`.
        """
        if vmin is None:
            vmin, _ = self._data.GetScalarRange()
        if vmax is None:
            _, vmax = self._data.GetScalarRange()
        self._alpha_grad = alpha_grad
        volumeProperty = self.GetProperty()

        if alpha_grad is None:
            volumeProperty.DisableGradientOpacityOn()
            return self

        volumeProperty.DisableGradientOpacityOff()

        gotf = volumeProperty.GetGradientOpacity()
        if utils.is_sequence(alpha_grad):
            alpha_grad = np.array(alpha_grad)
            if len(alpha_grad.shape) == 1:  # user passing a flat list e.g. (0.0, 0.3, 0.9, 1)
                for i, al in enumerate(alpha_grad):
                    xalpha = vmin + (vmax - vmin) * i / (len(alpha_grad) - 1)
                    # Create transfer mapping scalar value to gradient opacity
                    gotf.AddPoint(xalpha, al)
            elif len(alpha_grad.shape) == 2:  # user passing [(x0,alpha0), ...]
                gotf.AddPoint(vmin, alpha_grad[0][1])
                for xalpha, al in alpha_grad:
                    # Create transfer mapping scalar value to opacity
                    gotf.AddPoint(xalpha, al)
                gotf.AddPoint(vmax, alpha_grad[-1][1])
            # print("alpha_grad at", round(xalpha, 1), "\tset to", al)
        else:
            gotf.AddPoint(vmin, alpha_grad)  # constant alpha_grad
            gotf.AddPoint(vmax, alpha_grad)
        return self

    def component_weight(self, i, weight):
        """Set the scalar component weight in range [0,1]."""
        self.GetProperty().SetComponentWeight(i, weight)
        return self

    def xslice(self, i):
        """Extract the slice at index `i` of volume along x-axis."""
        vslice = vtk.vtkImageDataGeometryFilter()
        vslice.SetInputData(self.imagedata())
        nx, ny, nz = self.imagedata().GetDimensions()
        if i > nx - 1:
            i = nx - 1
        vslice.SetExtent(i, i, 0, ny, 0, nz)
        vslice.Update()
        m = Mesh(vslice.GetOutput())
        m.pipeline = utils.OperationNode(f"xslice {i}", parents=[self], c="#4cc9f0:#e9c46a")
        return m

    def yslice(self, j):
        """Extract the slice at index `j` of volume along y-axis."""
        vslice = vtk.vtkImageDataGeometryFilter()
        vslice.SetInputData(self.imagedata())
        nx, ny, nz = self.imagedata().GetDimensions()
        if j > ny - 1:
            j = ny - 1
        vslice.SetExtent(0, nx, j, j, 0, nz)
        vslice.Update()
        m = Mesh(vslice.GetOutput())
        m.pipeline = utils.OperationNode(f"yslice {j}", parents=[self], c="#4cc9f0:#e9c46a")
        return m

    def zslice(self, k):
        """Extract the slice at index `i` of volume along z-axis."""
        vslice = vtk.vtkImageDataGeometryFilter()
        vslice.SetInputData(self.imagedata())
        nx, ny, nz = self.imagedata().GetDimensions()
        if k > nz - 1:
            k = nz - 1
        vslice.SetExtent(0, nx, 0, ny, k, k)
        vslice.Update()
        m = Mesh(vslice.GetOutput())
        m.pipeline = utils.OperationNode(f"zslice {k}", parents=[self], c="#4cc9f0:#e9c46a")
        return m

    @deprecated(reason=vedo.colors.red + "Please use slice_plane()" + vedo.colors.reset)
    def slicePlane(self, *a, **b):
        "Deprecated. Please use `slice_plane()`"
        return self.slice_plane(*a, **b)

    def slice_plane(self, origin=(0, 0, 0), normal=(1, 1, 1), autocrop=False):
        """
        Extract the slice along a given plane position and normal.

        Example:
            - [slice_plane1.py](https://github.com/marcomusy/vedo/tree/master/examples/volumetric/slice_plane1.py)

                ![](https://vedo.embl.es/images/volumetric/slicePlane1.gif)
        """
        reslice = vtk.vtkImageReslice()
        reslice.SetInputData(self._data)
        reslice.SetOutputDimensionality(2)
        newaxis = utils.versor(normal)
        pos = np.array(origin)
        initaxis = (0, 0, 1)
        crossvec = np.cross(initaxis, newaxis)
        angle = np.arccos(np.dot(initaxis, newaxis))
        T = vtk.vtkTransform()
        T.PostMultiply()
        T.RotateWXYZ(np.rad2deg(angle), crossvec)
        T.Translate(pos)
        M = T.GetMatrix()
        reslice.SetResliceAxes(M)
        reslice.SetInterpolationModeToLinear()
        reslice.SetAutoCropOutput(not autocrop)
        reslice.Update()
        vslice = vtk.vtkImageDataGeometryFilter()
        vslice.SetInputData(reslice.GetOutput())
        vslice.Update()
        msh = Mesh(vslice.GetOutput())
        msh.SetOrientation(T.GetOrientation())
        msh.SetPosition(pos)
        msh.pipeline = utils.OperationNode("slice_plane", parents=[self], c="#4cc9f0:#e9c46a")
        return msh

    def warp(self, source, target, sigma=1, mode="3d", fit=False):
        """
        Warp volume scalars within a Volume by specifying
        source and target sets of points.

        Arguments:
            source : (Points, list)
                the list of source points
            target : (Points, list)
                the list of target points
            fit : (bool)
                fit/adapt the old bounding box to the warped geometry
        """
        if isinstance(source, vedo.Points):
            source = source.points()
        if isinstance(target, vedo.Points):
            target = target.points()

        ns = len(source)
        ptsou = vtk.vtkPoints()
        ptsou.SetNumberOfPoints(ns)
        for i in range(ns):
            ptsou.SetPoint(i, source[i])

        nt = len(target)
        if ns != nt:
            vedo.logger.error(f"#source {ns} != {nt} #target points")
            raise RuntimeError()

        pttar = vtk.vtkPoints()
        pttar.SetNumberOfPoints(nt)
        for i in range(ns):
            pttar.SetPoint(i, target[i])

        T = vtk.vtkThinPlateSplineTransform()
        if mode.lower() == "3d":
            T.SetBasisToR()
        elif mode.lower() == "2d":
            T.SetBasisToR2LogR()
        else:
            vedo.logger.error(f"unknown mode {mode}")
            raise RuntimeError()

        T.SetSigma(sigma)
        T.SetSourceLandmarks(ptsou)
        T.SetTargetLandmarks(pttar)
        T.Inverse()
        self.transform = T
        self.apply_transform(T, fit=fit)
        self.pipeline = utils.OperationNode("warp", parents=[self], c="#4cc9f0")
        return self

    def apply_transform(self, T, fit=False):
        """
        Apply a transform to the scalars in the volume.

        Arguments:
            T : (vtkTransform, matrix)
                The transformation to be applied
            fit : (bool)
                fit/adapt the old bounding box to the warped geometry
        """
        if isinstance(T, vtk.vtkMatrix4x4):
            tr = vtk.vtkTransform()
            tr.SetMatrix(T)
            T = tr

        elif utils.is_sequence(T):
            M = vtk.vtkMatrix4x4()
            n = len(T[0])
            for i in range(n):
                for j in range(n):
                    M.SetElement(i, j, T[i][j])
            tr = vtk.vtkTransform()
            tr.SetMatrix(M)
            T = tr

        reslice = vtk.vtkImageReslice()
        reslice.SetInputData(self._data)
        reslice.SetResliceTransform(T)
        reslice.SetOutputDimensionality(3)
        reslice.SetInterpolationModeToLinear()

        spacing = self._data.GetSpacing()
        origin = self._data.GetOrigin()

        if fit:
            bb = self.box()
            if isinstance(T, vtk.vtkThinPlateSplineTransform):
                TI = vtk.vtkThinPlateSplineTransform()
                TI.DeepCopy(T)
                TI.Inverse()
            else:
                TI = vtk.vtkTransform()
                TI.DeepCopy(T)
            bb.apply_transform(TI)
            bounds = bb.GetBounds()
            bounds = (
                bounds[0] / spacing[0],
                bounds[1] / spacing[0],
                bounds[2] / spacing[1],
                bounds[3] / spacing[1],
                bounds[4] / spacing[2],
                bounds[5] / spacing[2],
            )
            bounds = np.round(bounds).astype(int)
            reslice.SetOutputExtent(bounds)
            reslice.SetOutputSpacing(spacing[0], spacing[1], spacing[2])
            reslice.SetOutputOrigin(origin[0], origin[1], origin[2])

        reslice.Update()
        self._update(reslice.GetOutput())
        self.pipeline = utils.OperationNode("apply_transform", parents=[self], c="#4cc9f0")
        return self


##########################################################################
class VolumeSlice(BaseVolume, Base3DProp, vtk.vtkImageSlice):
    """
    Derived class of `vtkImageSlice`.
    """

    def __init__(self, inputobj=None):
        """
        This class is equivalent to `Volume` except for its representation.
        The main purpose of this class is to be used in conjunction with `Volume`
        for visualization using `mode="image"`.
        """
        vtk.vtkImageSlice.__init__(self)
        Base3DProp.__init__(self)
        BaseVolume.__init__(self)
        # super().__init__()

        self._mapper = vtk.vtkImageResliceMapper()
        self._mapper.SliceFacesCameraOn()
        self._mapper.SliceAtFocalPointOn()
        self._mapper.SetAutoAdjustImageQuality(False)
        self._mapper.BorderOff()

        self.lut = None

        self.property = vtk.vtkImageProperty()
        self.property.SetInterpolationTypeToLinear()
        self.SetProperty(self.property)

        ###################
        if isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = vedo.file_io.download(inputobj, verbose=False)  # fpath
            elif os.path.isfile(inputobj):
                pass
            else:
                inputobj = sorted(glob.glob(inputobj))

        ###################
        inputtype = str(type(inputobj))

        if inputobj is None:
            img = vtk.vtkImageData()

        if isinstance(inputobj, Volume):
            img = inputobj.imagedata()
            self.lut = utils.ctf2lut(inputobj)

        elif utils.is_sequence(inputobj):

            if isinstance(inputobj[0], str):  # scan sequence of BMP files
                ima = vtk.vtkImageAppend()
                ima.SetAppendAxis(2)
                pb = utils.ProgressBar(0, len(inputobj))
                for i in pb.range():
                    f = inputobj[i]
                    picr = vtk.vtkBMPReader()
                    picr.SetFileName(f)
                    picr.Update()
                    mgf = vtk.vtkImageMagnitude()
                    mgf.SetInputData(picr.GetOutput())
                    mgf.Update()
                    ima.AddInputData(mgf.GetOutput())
                    pb.print("loading...")
                ima.Update()
                img = ima.GetOutput()

            else:
                if "ndarray" not in inputtype:
                    inputobj = np.array(inputobj)

                if len(inputobj.shape) == 1:
                    varr = utils.numpy2vtk(inputobj, dtype=float)
                else:
                    if len(inputobj.shape) > 2:
                        inputobj = np.transpose(inputobj, axes=[2, 1, 0])
                    varr = utils.numpy2vtk(inputobj.ravel(order="F"), dtype=float)
                varr.SetName("input_scalars")

                img = vtk.vtkImageData()
                img.SetDimensions(inputobj.shape)
                img.GetPointData().AddArray(varr)
                img.GetPointData().SetActiveScalars(varr.GetName())

        elif "ImageData" in inputtype:
            img = inputobj

        elif isinstance(inputobj, Volume):
            img = inputobj.inputdata()

        elif "UniformGrid" in inputtype:
            img = inputobj

        elif hasattr(inputobj, "GetOutput"):  # passing vtk object, try extract imagdedata
            if hasattr(inputobj, "Update"):
                inputobj.Update()
            img = inputobj.GetOutput()

        elif isinstance(inputobj, str):
            if "https://" in inputobj:
                inputobj = vedo.file_io.download(inputobj, verbose=False)
            img = vedo.file_io.loadImageData(inputobj)

        else:
            vedo.logger.error(f"cannot understand input type {inputtype}")
            return

        self._data = img
        self._mapper.SetInputData(img)
        self.SetMapper(self._mapper)

    def bounds(self):
        """Return the bounding box as [x0,x1, y0,y1, z0,z1]"""
        bns = [0, 0, 0, 0, 0, 0]
        self.GetBounds(bns)
        return bns

    def colorize(self, lut=None, fix_scalar_range=False):
        """
        Assign a LUT (Look Up Table) to colorize the slice, leave it `None`
        to reuse an existing Volume color map.
        Use "bw" for automatic black and white.
        """
        if lut is None and self.lut:
            self.property.SetLookupTable(self.lut)
        elif isinstance(lut, vtk.vtkLookupTable):
            self.property.SetLookupTable(lut)
        elif lut == "bw":
            self.property.SetLookupTable(None)
        self.property.SetUseLookupTableScalarRange(fix_scalar_range)
        return self

    def alpha(self, value):
        """Set opacity to the slice"""
        self.property.SetOpacity(value)
        return self

    def auto_adjust_quality(self, value=True):
        """Automatically reduce the rendering quality for greater speed when interacting"""
        self._mapper.SetAutoAdjustImageQuality(value)
        return self

    def slab(self, thickness=0, mode=0, sample_factor=2):
        """
        Make a thick slice (slab).

        Arguments:
            thickness : (float)
                set the slab thickness, for thick slicing
            mode : (int)
                The slab type:
                    0 = min
                    1 = max
                    2 = mean
                    3 = sum
            sample_factor : (float)
                Set the number of slab samples to use as a factor of the number of input slices
                within the slab thickness. The default value is 2, but 1 will increase speed
                with very little loss of quality.
        """
        self._mapper.SetSlabThickness(thickness)
        self._mapper.SetSlabType(mode)
        self._mapper.SetSlabSampleFactor(sample_factor)
        return self

    def face_camera(self, value=True):
        """Make the slice always face the camera or not."""
        self._mapper.SetSliceFacesCameraOn(value)
        return self

    def jump_to_nearest_slice(self, value=True):
        """
        This causes the slicing to occur at the closest slice to the focal point,
        instead of the default behavior where a new slice is interpolated between
        the original slices.
        Nothing happens if the plane is oblique to the original slices."""
        self.SetJumpToNearestSlice(value)
        return self

    def fill_background(self, value=True):
        """
        Instead of rendering only to the image border,
        render out to the viewport boundary with the background color.
        The background color will be the lowest color on the lookup
        table that is being used for the image."""
        self._mapper.SetBackground(value)
        return self

    def lighting(self, window, level, ambient=1.0, diffuse=0.0):
        """Assign the values for window and color level."""
        self.property.SetColorWindow(window)
        self.property.SetColorLevel(level)
        self.property.SetAmbient(ambient)
        self.property.SetDiffuse(diffuse)
        return self



"""1D Moving Least Squares (MLS)
to project a cloud of unordered points
to become a smooth line"""
from vedo import *
import numpy as np

N = 4  # nr. of iterations

# build some initial cloud of noisy points along a line
pts = [ (sin(6*x), cos(2*x)*x, cos(9*x)) for x in np.arange(0,2, .001)]
# pts = [ (0, sin(x), cos(x)) for x in arange(0,6, .002) ]
# pts = [(sqrt(x), sin(x), x/5) for x in arange(0, 16, 0.01)]

pts += np.random.randn(len(pts), 3) /20  # add noise
np.random.shuffle(pts)  # make sure points are not ordered

pts = Points(pts, r=5)

plt = Plotter(N=N, axes=1)
plt.at(0).show(pts, __doc__)

for i in range(1, N):
    pts = pts.clone().smooth_mls_1d(f=0.4).color(i)

    if i == N-1:
        # at the last iteration make sure points
        # are separated by tol (in % of bbox)
        pts.subsample(0.02)

    plt.at(i).show(pts, f"Iteration {i}, #points: {pts.npoints}")

plt.interactive().close()



"""Cut a mesh with an other mesh and cap the holes"""
from vedo import dataurl, Plotter, Mesh, Sphere

msh1 = Mesh(dataurl+'motor.byu')
cutmesh = Sphere().y(-0.4).scale(0.4).wireframe().alpha(0.1)

msh2 = msh1.clone().cut_with_mesh(cutmesh)
redcap = msh2.cap(return_cap=True).color("r4")

plt = Plotter(N=2, axes=1)
plt.at(0).show(msh1, cutmesh, __doc__)
plt.at(1).show(msh2, redcap, viewup="z")
plt.interactive().close()



"""Interpolate cell values from a quad-mesh to a tri-mesh"""
from vedo import Grid, show

# Make up some quad mesh with associated scalars
g1 = Grid(res=(25,25)).wireframe(0).lw(1)
scalars = g1.points()[:,1]
g1.cmap("viridis", scalars, vmin=-1, vmax=1, name='gene')
g1.map_points_to_cells()  # move the array to cells (faces)
g1.add_scalarbar(horizontal=1, pos=(0.7,0.04))
g1.rotate_z(20)  # let's rotate it a bit so it's visible

# Interpolate first mesh onto a new triangular mesh
eps = 0.01
g2 = Grid(res=(50,50)).pos(0.2, 0.2, 0.1).wireframe(0).lw(0)
g2.triangulate()

# Interpolate by averaging the closest 3 points:
#g2.interpolate_data_from(g1, on='cells', n=3)

# Interpolate by picking points in a specified radius,
#  if there are no points in that radius set null value -1
g2.interpolate_data_from(
    g1,
    on='cells',
    radius=0.1+eps,
    null_strategy=1,
    null_value=-1,
)

g2.cmap('hot', 'gene', on='cells', vmin=-1, vmax=1).add_scalarbar()

show(g1, g2, __doc__, axes=1)



"""Example usage of remove_outliers() and compute_clustering()"""
# Import the vedo library and numpy
from vedo import np, Points, show

# Generate 4 random sets of N points in 3D space
N = 2000
f = 0.6
noise1 = np.random.rand(N, 3) * f + np.array([1, 1, 0])
noise2 = np.random.rand(N, 3) * f + np.array([1, 0, 1.2])
noise3 = np.random.rand(N, 3) * f + np.array([0, 1, 1])
noise4 = np.random.randn(N, 3) * f / 8 + np.array([1, 1, 1])

# Create a Points object from the noisy point sets
noise4 = Points(noise4).remove_outliers(radius=0.05).points()
pts = noise1.tolist() + noise2.tolist() + noise3.tolist() + noise4.tolist()
pts = Points(pts)

# Cluster the points to find back their original identity
clpts = pts.compute_clustering(radius=0.1).print()
# Set the color of the points based on their cluster ID using the 'jet' colormap
clpts.cmap("jet", "ClusterId")

show(clpts, __doc__, axes=1, viewup='z', bg='blackboard').close()
